import itertools
from typing import Iterable, Optional, Tuple


def sequential_pair_selector(
    n_sources: int, stop: Optional[int] = None, step: int = 1, sort: bool = False
) -> Iterable[Tuple[int, int]]:
    r"""Select pair in pairwise update.

    Args:
        n_sources (int):
            Number of sources.
        step (int):
            This parameter determines step size.
            For instance, if ``sequential_pair_selector(n_sources=6, step=2, sort=False)``,
            this function yields ``0, 1``, ``2, 3``, ``4, 5``, ``0, 1``, ``2, 3``, ``4, 5``.
            Default: ``1``.
        sort (bool):
            Sort pair to ensure :math:`m<n` if ``sort=True``.
            Default: ``False``.

    Yields:
        Pair (tuple) of indices.

    Examples:
        .. code-block:: python

            >>> for m, n in combination_pair_selector(4):
            ...     print(m, n)
            0 1
            1 2
            2 3
            3 0
    """
    if stop is None:
        stop = n_sources

    for m in range(0, stop, step):
        m, n = m % n_sources, (m + 1) % n_sources

        if sort:
            m, n = (n, m) if m > n else (m, n)

        yield m, n


def combination_pair_selector(n_sources: int, sort: bool = False) -> Iterable[Tuple[int, int]]:
    r"""Select pair in pairwise update.

    Args:
        n_sources (int):
            Number of sources.
        sort (bool):
            Sort pair to ensure :math:`m<n` if ``sort=True``.
            Default: ``False``.

    Yields:
        Pair (tuple) of indices.

    Examples:
        .. code-block:: python

            >>> for m, n in combination_pair_selector(4):
            ...     print(m, n)
            0 1
            0 2
            0 3
            1 2
            1 3
            2 3
    """
    for m, n in itertools.combinations(range(n_sources), 2):
        if sort:
            m, n = (n, m) if m > n else (m, n)

        yield m, n



import functools
import math
from typing import Callable, List, Optional, Union

import numpy as np

from ..special.flooring import identity, max_flooring
from .pdsbss import MaskingPDSBSS

__all__ = [
    "MaskingPDSHVA",
    "HVA",
]

EPS = 1e-10


class MaskingPDSHVA(MaskingPDSBSS):
    r"""Harmonic vector analysis proposed in [#yatabe2021determined]_.

    Args:
        mu1 (float):
            Step size. Default: ``1``.
        mu2 (float):
            Step size. Default: ``1``.
        alpha (float):
            Relaxation parameter (deprecated). Set ``relaxation`` instead.
        relaxation (float):
            Relaxation parameter. Default: ``1``.
        attenuation (float, optional):
            Attenuation parameter in masking. Default: ``1 / n_sources``.
        mask_iter (int):
            Number of iterations in application of cosine shrinkage operator.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.

    .. [#yatabe2021determined] K. Yatabe and D. Kitamura,
        "Determined BSS based on time-frequency masking and its application to \
        harmonic vector analysis," *IEEE/ACM Trans. ASLP*, vol. 29, pp. 1609-1625, 2021.

    """

    def __init__(
        self,
        mu1: float = 1,
        mu2: float = 1,
        alpha: float = None,
        relaxation: float = 1,
        attenuation: Optional[float] = None,
        mask_iter: int = 1,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[Callable[["MaskingPDSHVA"], None], List[Callable[["MaskingPDSHVA"], None]]]
        ] = None,
        scale_restoration: bool = True,
        record_loss: Optional[bool] = None,
        reference_id: int = 0,
    ) -> None:
        def mask_fn(y: np.ndarray) -> np.ndarray:
            """Masking function to emphasize harmonic components.

            Args:
                y (np.ndarray):
                    The shape is (n_sources, n_bins, n_frames).

            Returns:
                np.ndarray of mask. The shape is (n_sources, n_bins, n_frames).
            """
            n_sources, n_bins, _ = y.shape

            if self.attenuation is None:
                self.attenuation = 1 / n_sources

            gamma = self.attenuation

            y = self.flooring_fn(np.abs(y))
            zeta = np.log(y)
            zeta_mean = zeta.mean(axis=1, keepdims=True)
            rho = zeta - zeta_mean
            nu = np.fft.irfft(rho, axis=1, norm="backward")
            nu = nu[:, :n_bins]
            varsigma = np.minimum(1, nu)

            for _ in range(mask_iter):
                varsigma = (1 - np.cos(math.pi * varsigma)) / 2

            xi = np.fft.irfft(varsigma * nu, axis=1, norm="forward")
            xi = xi[:, :n_bins]
            varrho = xi + zeta_mean
            v = np.exp(2 * varrho)
            mask = (v / v.sum(axis=0)) ** gamma

            return mask

        super().__init__(
            mu1=mu1,
            mu2=mu2,
            alpha=alpha,
            relaxation=relaxation,
            penalty_fn=None,
            mask_fn=mask_fn,
            callbacks=callbacks,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
        )

        self.attenuation = attenuation
        self.mask_iter = mask_iter

        if flooring_fn is None:
            self.flooring_fn = identity
        else:
            self.flooring_fn = flooring_fn

    def __repr__(self) -> str:
        s = "MaskingPDSHVA("
        s += "mu1={mu1}, mu2={mu2}"
        s += ", relaxation={relaxation}"

        if self.attenuation is not None:
            s += ", attenuation={attenuation}"

        s += ", mask_iter={mask_iter}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)


class HVA(MaskingPDSHVA):
    """Alias of MaskingPDSHVA."""

    def __repr__(self) -> str:
        s = "HVA("
        s += "mu1={mu1}, mu2={mu2}"
        s += ", relaxation={relaxation}"

        if self.attenuation is not None:
            s += ", attenuation={attenuation}"

        s += ", mask_iter={mask_iter}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)



try:
    from .io import wavread, wavwrite
except ModuleNotFoundError:
    # to avoid module not found error during installation
    # e.g. numpy is not found in io.py
    pass

try:
    from ._version import __version__
except ModuleNotFoundError:
    __version__ = "0.1.7"

__all__ = ["__version__", "wavread", "wavwrite"]



import os
import shutil
import urllib.request

import numpy as np

from ...io import wavread


def download(root: str = ".data/SiSEC2010", n_sources: int = 3, tag: str = "dev1_female3") -> str:
    filename = "dev1.zip"
    url = "http://www.irisa.fr/metiss/SiSEC10/underdetermined/{}".format(filename)
    zip_path = os.path.join(root, filename)

    os.makedirs(root, exist_ok=True)

    if not os.path.exists(zip_path):
        urllib.request.urlretrieve(url, zip_path)

    if not os.path.exists(os.path.join(root, "{}_inst_matrix.mat".format(tag))):
        shutil.unpack_archive(zip_path, root)

    source_paths = []

    for src_idx in range(n_sources):
        source_path = os.path.join(root, "{}_src_{}.wav".format(tag, src_idx + 1))
        source_paths.append(source_path)

    channels = [3, 4, 2, 5]
    sample_rate = 16000

    source_paths = source_paths[:n_sources]
    channels = channels[:n_sources]

    n_channels = len(channels)
    npz_path = os.path.join(root, "SiSEC2010-{}ch.npz".format(n_channels))

    assert n_channels == n_sources, "Mixing system should be determined."

    if not os.path.exists(npz_path):
        dry_sources = {}

        for src_idx, source_path in enumerate(source_paths):
            data, _ = wavread(source_path, return_2d=False)
            dry_sources["src_{}".format(src_idx + 1)] = data

        np.savez(
            npz_path,
            sample_rate=sample_rate,
            n_sources=n_sources,
            n_channels=n_channels,
            **dry_sources,
        )

    return npz_path



import functools
import itertools
from typing import Callable, Optional

import numpy as np

from ..special.flooring import identity, max_flooring

EPS = 1e-10


def correlation_based_permutation_solver(
    sequence: np.ndarray,
    *args,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
    overwrite: bool = True,
) -> np.ndarray:
    r"""Solve permutation of estimated spectrograms.

    Group channels at each frequency bin according to correlations
    between frequencies [#murata2001approach]_.

    Args:
        sequence (numpy.ndarray):
            Array-like sequence of shape (n_bins, n_sources, n_frames).
        args (tuple of numpy.ndarray, optional):
            Positional arguments each of which is ``numpy.ndarray``.
            The shapes of each item should be (n_bins, n_sources, \*).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``partial(max_flooring, eps=1e-10)``.
        overwrite (bool):
            Overwrite ``sequence`` and ``args`` if ``overwrite=True``.
            Default: ``True``.

    Returns:
        - If ``args`` is not given, ``numpy.ndarray`` of permutated separated spectrograms
            with shape of (n_sources, n_bins, n_frames) are returned.
        - If one positional argument is given, ``numpy.ndarray``s of permutated separated
            spectrograms and the permutated positional argument are returned.
        - If more than two positional arguments are given, ``numpy.ndarray``s of
            permutated separated spectrograms and the permutated positional arguments are returned.

        .. [#murata2001approach]
            N. Murata, S. Ikeda, and A. Ziehe,
            "An approach to blind source separation based on temporal structure of speech signals,"
            in *Neurocomputing*, vol. 41, no. 1, pp. 1-24, 2001.

    .. note::

        In this function, the shape of ``separated`` is expected ``(n_bins, n_sources, ...)``,
        which is different from other functions.
    """
    assert sequence.ndim == 3, "Dimension of sequence is expected to be 3."

    for pos_idx, arg in enumerate(args):
        if arg.shape[:2] != sequence.shape[:2]:
            raise ValueError("The shape of {}th argument is invalid.".format(pos_idx + 1))

    if overwrite:
        Y = sequence
        permutable = args
    else:
        Y = sequence.copy()

        permutable = []

        for arg in args:
            permutable.append(arg.copy())

        permutable = tuple(permutable)

    if flooring_fn is None:
        flooring_fn = identity
    else:
        flooring_fn = flooring_fn

    n_bins, n_sources, _ = Y.shape

    permutations = list(itertools.permutations(range(n_sources)))

    P = np.abs(Y)
    norm = np.sqrt(np.sum(P**2, axis=1, keepdims=True))
    norm = flooring_fn(norm)
    P = P / norm
    correlation = np.sum(P @ P.transpose(0, 2, 1), axis=(1, 2))
    indices = np.argsort(correlation)

    min_idx = indices[0]
    P_criteria = P[min_idx]

    for bin_idx in range(1, n_bins):
        min_idx = indices[bin_idx]
        P_max = None
        perm_max = None

        for perm in permutations:
            P_perm = np.sum(P_criteria * P[min_idx, perm, :])

            if P_max is None or P_perm > P_max:
                P_max = P_perm
                perm_max = perm

        P_criteria = P_criteria + P[min_idx, perm_max, :]
        Y[min_idx, :] = Y[min_idx, perm_max]

        for idx in range(len(permutable)):
            permutable[idx][min_idx, :] = permutable[idx][min_idx, perm_max]

    if len(permutable) == 0:
        return Y
    elif len(permutable) == 1:
        return Y, permutable[0]
    else:
        return Y, permutable


def score_based_permutation_solver(
    sequence: np.ndarray,
    *args,
    global_iter: int = 1,
    local_iter: int = 1,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
    multi_centroids: bool = False,
    overwrite: bool = True,
) -> np.ndarray:
    r"""Align permutations between frequencies based on score value [#sawada2010underdetermined]_.

    Args:
        sequence (numpy.ndarray):
            Array-like sequence of shape (n_bins, n_sources, n_frames).
        args (tuple of numpy.ndarray, optional):
            Positional arguments each of which is ``numpy.ndarray``.
            The shapes of each item should be (n_bins, n_sources, \*).
        global_iter (int):
            Number of iterations in global optimization.
        local_iter (int):
            Number of iterations in local optimization.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to receive (n_channels, n_bins, n_frames)
            and return (n_channels, n_bins, n_frames).
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``partial(max_flooring, eps=1e-10)``.
        multi_centroids (bool):
            If ``multi_centroids=True``, multiple centroids are used in global optimization.
            However, this is not supported now. Default: ``False``.
        overwrite (bool):
            Overwrite ``sequence`` and ``args`` if ``overwrite=True``.
            Default: ``True``.

    .. [#sawada2010underdetermined]
        H. Sawada, S. Araki, and S. Makino,
        "Underdetermined convolutive blind source separation \
        via frequency bin-wise clustering and permutation alignment,"
        in *IEEE Trans. ASLP*, vol. 19, no. 3, pp. 516-527, 2010.
    """
    assert sequence.ndim == 3, "Dimension of sequence is expected to be 3."
    assert not multi_centroids, "multi_centroids version is not supported."

    for pos_idx, arg in enumerate(args):
        if arg.shape[:2] != sequence.shape[:2]:
            raise ValueError("The shape of {}th argument is invalid.".format(pos_idx + 1))

    if overwrite:
        permutable = args
    else:
        sequence = sequence.copy()

        permutable = []

        for arg in args:
            permutable.append(arg.copy())

        permutable = tuple(permutable)

    if flooring_fn is None:
        flooring_fn = identity
    else:
        flooring_fn = flooring_fn

    n_bins, n_sources = sequence.shape[:2]
    na = np.newaxis
    eye = np.eye(n_sources)
    permutations = np.array(list(itertools.permutations(range(n_sources))))

    sequence_mean = sequence.mean(axis=-1, keepdims=True)
    sequence_std = sequence.std(axis=-1, keepdims=True)
    sequence_normalized = (sequence - sequence_mean) / sequence_std

    for _ in range(global_iter):
        centroid = sequence_normalized.mean(axis=0)
        centroid_std = centroid.std(axis=-1, keepdims=True)
        scores = []

        for perm in permutations:
            num = np.mean(sequence_normalized[:, perm, na] * centroid[na, :], axis=-1)
            denom = flooring_fn(centroid_std)
            corr = num / denom
            score = np.sum(eye * corr - (1 - eye) * corr, axis=(1, 2))
            scores.append(score)

        scores = np.stack(scores, axis=1)
        perm_max = np.argmax(scores, axis=1)
        perm_max = permutations[perm_max]
        sequence_normalized = _parallel_sort(sequence_normalized, perm_max)
        sequence = _parallel_sort(sequence, perm_max)

        for idx in range(len(permutable)):
            permutable[idx][:] = _parallel_sort(permutable[idx], perm_max)

    # local optimization
    for _ in range(local_iter):
        for bin_idx in range(n_bins):
            min_idx = max(0, bin_idx - 3)
            max_idx = min(n_bins - 1, bin_idx + 3)
            covariant_indices = set(range(min_idx, bin_idx)) | set(range(bin_idx + 1, max_idx + 1))

            min_idx = max(0, bin_idx // 2 - 1)
            max_idx = min(n_bins - 1, bin_idx // 2 + 1)
            covariant_indices |= set(range(min_idx, max_idx + 1))

            min_idx = max(0, 2 * bin_idx - 1)
            max_idx = min(n_bins - 1, 2 * bin_idx + 1)
            covariant_indices |= set(range(min_idx, max_idx + 1))

            # deterministic
            covariant_indices = sorted(list(covariant_indices))
            covariant_sequence = sequence_normalized[covariant_indices]

            scores = []

            for perm in permutations:
                num = np.mean(
                    sequence_normalized[bin_idx, perm, na] * covariant_sequence[:, na],
                    axis=-1,
                )
                denom = flooring_fn(centroid_std)
                corr = num / denom
                score = np.sum(eye * corr - (1 - eye) * corr, axis=(1, 2))
                score = score.sum(axis=0)
                scores.append(score)

            scores = np.stack(scores, axis=0)
            perm_max = np.argmax(scores, axis=0)
            perm_max = permutations[perm_max]
            sequence_normalized[bin_idx] = sequence_normalized[bin_idx, perm_max]
            sequence[bin_idx] = sequence[bin_idx, perm_max]

            for idx in range(len(permutable)):
                permutable[idx][bin_idx] = permutable[idx][bin_idx, perm_max]

    if len(permutable) == 0:
        return sequence
    elif len(permutable) == 1:
        return sequence, permutable[0]
    else:
        return sequence, permutable


def _parallel_sort(X: np.ndarray, indices: np.ndarray) -> np.ndarray:
    shape = X.shape
    idx = np.repeat(indices, repeats=np.prod(shape[2:]), axis=-1).reshape(shape)
    X = np.take_along_axis(X, idx, axis=1)

    return X



import numpy as np


def pca(input: np.ndarray, ascend: bool = True) -> np.ndarray:
    r"""Apply principal component analysis (PCA).

    Args:
        input (numpy.ndarray):
            Input tensor.
        ascend (bool):
            If ``ascend=True``, first channel corresponds to first principle component. \
            Otherwise, last channel corresponds to first principle component.

    Returns:
        numpy.ndarray:
            Output tensor. The type (real or complex) and shape are same as input.

    .. note::
        - If ``input`` is 2D real tensor, it is regarded as (n_channels, n_samples).
        - If ``input`` is 3D complex tensor, it is regarded as (n_channels, n_bins, n_frames).
        - If ``input`` is 3D real tensor, it is regarded as (batch_size, n_channels, n_samples).
        - If ``input`` is 4D complex tensor, it is regarded as \
          (batch_size, n_channels, n_bins, n_frames).

    Examples:
        .. code-block:: python

            >>> import numpy as np
            >>> from ssspy.transform import pca

            >>> n_channels, n_bins, n_frames = 2, 2049, 128
            >>> n_sources = n_channels
            >>> rng = np.random.default_rng(42)

            >>> spectrogram_mix = \
            ...     rng.standard_normal((n_channels, n_bins, n_frames)) \
            ...     + 1j * rng.standard_normal((n_channels, n_bins, n_frames))
            >>> spectrogram_mix_ortho = pca(spectrogram_mix)
            >>> spectrogram_mix_ortho.shape
            (2, 2049, 128)
    """
    if input.ndim == 2:
        if np.iscomplexobj(input):
            raise ValueError("Real tensor is expected, but given complex tensor.")
        else:
            X = input.transpose(1, 0)
            covariance = np.mean(X[:, :, np.newaxis] * X[:, np.newaxis, :], axis=0)
            _, V = np.linalg.eigh(covariance)

            if ascend:
                V = V[..., ::-1]

            Y = X @ V
            output = Y.transpose(1, 0)
    elif input.ndim == 3:
        if np.iscomplexobj(input):
            X = input.transpose(1, 2, 0)
            covariance = np.mean(X[:, :, :, np.newaxis] * X[:, :, np.newaxis, :].conj(), axis=1)
            _, V = np.linalg.eigh(covariance)

            if ascend:
                V = V[..., ::-1]

            Y = X @ V.conj()
            output = Y.transpose(2, 0, 1)
        else:
            X = input.transpose(0, 2, 1)
            covariance = np.mean(X[:, :, :, np.newaxis] * X[:, :, np.newaxis, :], axis=1)
            _, V = np.linalg.eigh(covariance)

            if ascend:
                V = V[..., ::-1]

            Y = X @ V
            output = Y.transpose(0, 2, 1)
    elif input.ndim == 4:
        if np.iscomplexobj(input):
            X = input.transpose(0, 2, 3, 1)
            covariance = np.mean(
                X[:, :, :, :, np.newaxis] * X[:, :, :, np.newaxis, :].conj(), axis=2
            )
            _, V = np.linalg.eigh(covariance)

            if ascend:
                V = V[..., ::-1]

            Y = X @ V.conj()
            output = Y.transpose(0, 3, 1, 2)
        else:
            raise ValueError("Complex tensor is expected, but given real tensor.")
    else:
        raise ValueError(
            "The dimension of input is expected 3 or 4, but given {}.".format(input.ndim)
        )

    return output



import functools
from typing import Callable, Iterable, List, Optional, Tuple, Union

import numpy as np

from ..linalg.mean import gmeanmh
from ..special.flooring import identity, max_flooring
from ..special.psd import to_psd
from ..utils.flooring import choose_flooring_fn
from ..utils.select_pair import sequential_pair_selector
from ._update_spatial_model import update_by_ip1, update_by_ip2
from .base import IterativeMethodBase

__all__ = ["GaussMNMF", "FastGaussMNMF"]

diagonalizer_algorithms = ["IP", "IP1", "IP2"]
EPS = 1e-10


class MNMFBase(IterativeMethodBase):
    r"""Base class of multichannel nonnegative matrix factorization (MNMF).

    Args:
        n_basis (int):
            Number of NMF bases.
        n_sources (int, optional):
            Number of sources to be separated.
            If ``None`` is given, ``n_sources`` is determined by number of channels
            in input spectrogram. Default: ``None``.
        partitioning (bool):
            Whether to use partioning function. Default: ``False``.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel in multichannel Wiener filter. Default: ``0``.
        rng (numpy.random.Generator, optioinal):
            Random number generator. This is mainly used to randomly initialize PSDTF.
            If ``None`` is given, ``np.random.default_rng()`` is used.
            Default: ``None``.
    """

    def __init__(
        self,
        n_basis: int,
        n_sources: Optional[int] = None,
        partitioning: bool = False,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[Callable[["MNMFBase"], None], List[Callable[["MNMFBase"], None]]]
        ] = None,
        normalization: bool = True,
        record_loss: bool = True,
        reference_id: int = 0,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        super().__init__(callbacks=callbacks, record_loss=record_loss)

        self.n_basis = n_basis
        self.n_sources = n_sources
        self.partitioning = partitioning

        if flooring_fn is None:
            self.flooring_fn = identity
        else:
            self.flooring_fn = flooring_fn

        self.normalization = normalization

        self.input = None
        self.reference_id = reference_id

        if rng is None:
            rng = np.random.default_rng()

        self.rng = rng

    def __call__(
        self, input: np.ndarray, n_iter: int = 100, initial_call: bool = True, **kwargs
    ) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                The number of iterations of demixing filter updates.
                Default: ``100``.
            initial_call (bool):
                If ``True``, perform callbacks (and computation of loss if necessary)
                before iterations.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        super().__call__(n_iter=n_iter, initial_call=initial_call)

        self.output = self.separate(self.input)

        return self.output

    def __repr__(self) -> str:
        s = "MNMF("
        s += "n_basis={n_basis}"

        if self.n_sources is not None:
            s += ", n_sources={n_sources}"

        if hasattr(self, "n_channels"):
            s += ", n_channels={n_channels}"

        s += ", partitioning={partitioning}"
        s += ", normalization={normalization}"
        s += ", record_loss={record_loss}"
        s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(self, **kwargs) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            kwargs:
                Keyword arguments to set as attributes of MNMF.
        """
        assert self.input is not None, "Specify data!"

        for key in kwargs.keys():
            setattr(self, key, kwargs[key])

        X = self.input

        n_sources = self.n_sources
        n_channels, n_bins, n_frames = X.shape

        if n_sources is None:
            n_sources = n_channels

        self.n_sources, self.n_channels = n_sources, n_channels
        self.n_bins, self.n_frames = n_bins, n_frames

        self._init_instant_covariance()
        self._init_nmf(rng=self.rng)

        self.output = self.separate(X)

    def _init_instant_covariance(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Initialize instantaneous covariance of input.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        X = self.input
        XX = X[:, np.newaxis] * X[np.newaxis, :].conj()
        XX = XX.transpose(2, 3, 0, 1)
        self.instant_covariance = to_psd(XX, flooring_fn=flooring_fn)

    def _init_nmf(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        r"""Initialize NMF.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.
            rng (numpy.random.Generator, optional):
                Random number generator. If ``None`` is given,
                ``np.random.default_rng()`` is used.
                Default: ``None``.
        """
        n_basis = self.n_basis
        n_sources = self.n_sources
        n_bins, n_frames = self.n_bins, self.n_frames

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        if rng is None:
            rng = np.random.default_rng()

        if self.partitioning:
            if not hasattr(self, "basis"):
                T = rng.random((n_bins, n_basis))
                T = flooring_fn(T)
            else:
                # To avoid overwriting.
                T = self.basis.copy()

            if not hasattr(self, "activation"):
                V = rng.random((n_basis, n_frames))
                V = flooring_fn(V)
            else:
                # To avoid overwriting.
                V = self.activation.copy()

            if not hasattr(self, "latent"):
                Z = rng.random((n_sources, n_basis))
                Z = Z / Z.sum(axis=0)
                Z = flooring_fn(Z)
            else:
                # To avoid overwriting.
                Z = self.latent.copy()

            self.basis, self.activation = T, V
            self.latent = Z
        else:
            if not hasattr(self, "basis"):
                T = rng.random((n_sources, n_bins, n_basis))
                T = flooring_fn(T)
            else:
                # To avoid overwriting.
                T = self.basis.copy()

            if not hasattr(self, "activation"):
                V = rng.random((n_sources, n_basis, n_frames))
                V = flooring_fn(V)
            else:
                # To avoid overwriting.
                V = self.activation.copy()

            self.basis, self.activation = T, V

    def separate(self, input: np.ndarray) -> np.ndarray:
        raise NotImplementedError("Implement 'separate' method.")

    def reconstruct_nmf(
        self,
        basis: np.ndarray,
        activation: np.ndarray,
        latent: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""Reconstruct single-channel NMF.

        Args:
            basis (numpy.ndarray):
                Basis matrix.
                The shape is (n_sources, n_basis, n_frames) if latent is given.
                Otherwise, (n_basis, n_frames).
            activation (numpy.ndarray):
                Activation matrix.
                The shape is (n_sources, n_bins, n_basis) if latent is given.
                Otherwise, (n_bins, n_basis).
            latent (numpy.ndarray, optional):
                Latent variable that determines number of bases per source.

        Returns:
            numpy.ndarray of reconstructed single-channel NMF.
            The shape is (n_sources, n_bins, n_frames).
        """
        if latent is None:
            T, V = basis, activation
            Lamb = T @ V
        else:
            Z = latent
            T, V = basis, activation
            TV = T[:, :, np.newaxis] * V[np.newaxis, :, :]
            Lamb = np.sum(Z[:, np.newaxis, :, np.newaxis] * TV[np.newaxis, :, :, :], axis=2)

        return Lamb


class MNMF(MNMFBase):
    def __init__(
        self,
        n_basis: int,
        n_sources: Optional[int] = None,
        partitioning: bool = False,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[Union[Callable[["MNMF"], None], List[Callable[["MNMF"], None]]]] = None,
        normalization: bool = True,
        record_loss: bool = True,
        reference_id: int = 0,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        super().__init__(
            n_basis,
            n_sources=n_sources,
            partitioning=partitioning,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            normalization=normalization,
            record_loss=record_loss,
            reference_id=reference_id,
            rng=rng,
        )

    def _init_nmf(self, rng: Optional[np.random.Generator] = None) -> None:
        r"""Initialize NMF.

        Args:
            rng (numpy.random.Generator, optional):
                Random number generator. If ``None`` is given,
                ``np.random.default_rng()`` is used.
                Default: ``None``.
        """
        if rng is None:
            rng = np.random.default_rng()

        super()._init_nmf(rng=rng)

        n_sources, n_channels = self.n_sources, self.n_channels
        n_bins = self.n_bins

        if not hasattr(self, "spatial"):
            H = np.eye(n_channels, dtype=self.input.dtype)
            trace = np.trace(H, axis1=-2, axis2=-1)
            H = H / np.real(trace)
            H = np.tile(H, reps=(n_sources, n_bins, 1, 1))
        else:
            # To avoid overwriting.
            H = self.spatial.copy()

        self.spatial = H

    def reconstruct_mnmf(
        self,
        basis: np.ndarray,
        activation: np.ndarray,
        spatial: np.ndarray,
        latent: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""Reconstruct multichannel NMF.

        Args:
            basis (numpy.ndarray):
                Basis matrix with shape of (n_bins, n_basis).
            activation (numpy.ndarray):
                Activation matrix with shape of (n_basis, n_frames).
            spatial (numpy.ndarray):
                Spatial property with shape of (n_sources, n_bins, n_channels, n_channels).
            latent (numpy.ndarray, optional):
                Latent variables with shape of (n_sources, n_basis).

        Returns:
            numpy.ndarray of reconstructed multichannel NMF.
            The shape is (n_bins, n_frames, n_channels, n_channels).
        """
        T, V = basis, activation
        H = spatial

        if latent is None:
            Lamb = self.reconstruct_nmf(T, V)
        else:
            Lamb = self.reconstruct_nmf(T, V, latent=latent)

        R_n = Lamb[:, :, :, np.newaxis, np.newaxis] * H[:, :, np.newaxis, :, :]
        R = np.sum(R_n, axis=0)

        return R

    def normalize(self, axis1=-2, axis2=-1) -> None:
        r"""Ensure unit trace of spatial property of MNMF."""
        H = self.spatial
        n_dims = H.ndim

        axis1 = n_dims + axis1 if axis1 < 0 else axis1
        axis2 = n_dims + axis2 if axis2 < 0 else axis2

        assert axis1 == 2 and axis2 == 3

        trace = np.trace(H, axis1=axis1, axis2=axis2)
        trace = np.real(trace)
        H = H / trace[..., np.newaxis, np.newaxis]

        if self.partitioning:
            # When self.partitioning=True,
            # normalization may change value of cost function
            pass
        else:
            T = self.basis
            T = trace[:, :, np.newaxis] * T
            self.basis = T

        self.spatial = H


class FastMNMFBase(MNMFBase):
    r"""Base class of fast multichannel nonnegative matrix factorization (Fast MNMF).

    Args:
        n_basis (int):
            Number of NMF bases.
        n_sources (int, optional):
            Number of sources to be separated.
            If ``None`` is given, ``n_sources`` is determined by number of channels
            in input spectrogram. Default: ``None``.
        partitioning (bool):
            Whether to use partioning function. Default: ``False``.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        normalization (bool or str):
            Normalization of diagonalizers and diagonal elements of spatial covariance matrices.
            Only power-based normalization is supported. Default: ``True``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel in multichannel Wiener filter. Default: ``0``.
        rng (numpy.random.Generator, optioinal):
            Random number generator. This is mainly used to randomly initialize PSDTF.
            If ``None`` is given, ``np.random.default_rng()`` is used.
            Default: ``None``.
    """

    def __init__(
        self,
        n_basis: int,
        n_sources: Optional[int] = None,
        partitioning: bool = False,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[Callable[["FastMNMFBase"], None], List[Callable[["FastMNMFBase"], None]]]
        ] = None,
        normalization: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        super().__init__(
            n_basis,
            n_sources=n_sources,
            partitioning=partitioning,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            normalization=normalization,
            record_loss=record_loss,
            reference_id=reference_id,
            rng=rng,
        )

    def __repr__(self) -> str:
        s = "FastMNMF("
        s += "n_basis={n_basis}"

        if self.n_sources is not None:
            s += ", n_sources={n_sources}"

        if hasattr(self, "n_channels"):
            s += ", n_channels={n_channels}"

        s += ", partitioning={partitioning}"
        s += ", normalization={normalization}"
        s += ", record_loss={record_loss}"
        s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
        **kwargs,
    ) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.
            kwargs:
                Keyword arguments to set as attributes of MNMF.
        """
        assert self.input is not None, "Specify data!"

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        for key in kwargs.keys():
            setattr(self, key, kwargs[key])

        X = self.input

        n_sources = self.n_sources
        n_channels, n_bins, n_frames = X.shape

        if n_sources is None:
            n_sources = n_channels

        self.n_sources, self.n_channels = n_sources, n_channels
        self.n_bins, self.n_frames = n_bins, n_frames

        self._init_instant_covariance(flooring_fn=flooring_fn)
        self._init_nmf(flooring_fn=flooring_fn, rng=self.rng)
        self._init_diagonalizer(rng=self.rng)
        self._init_spatial(flooring_fn=flooring_fn, rng=self.rng)

        self.output = self.separate(X)

    def _init_diagonalizer(self, rng: Optional[np.random.Generator] = None) -> None:
        """Initialize diagonalizer.

        Args:
            rng (numpy.random.Generator, optional):
                Random number generator. If ``None`` is given,
                ``np.random.default_rng()`` is used.
                Default: ``None``.
        """
        n_channels = self.n_channels
        n_bins = self.n_bins

        if rng is None:
            rng = np.random.default_rng()

        if not hasattr(self, "diagonalizer"):
            Q = np.eye(n_channels, dtype=np.complex128)
            Q = np.tile(Q, reps=(n_bins, 1, 1))
        else:
            # To avoid overwriting.
            Q = self.diagonalizer.copy()

        self.diagonalizer = Q

    def _init_spatial(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        """Initialize diagonal elements of spatial covariance matrices.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.
            rng (numpy.random.Generator, optional):
                Random number generator. If ``None`` is given,
                ``np.random.default_rng()`` is used.
                Default: ``None``.
        """
        n_sources, n_channels = self.n_sources, self.n_channels
        n_bins = self.n_bins

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        if rng is None:
            rng = np.random.default_rng()

        if not hasattr(self, "spatial"):
            D = rng.random((n_bins, n_sources, n_channels))
            D = flooring_fn(D)
        else:
            D = self.spatial

        self.spatial = D

    def normalize(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Normalize diagonalizers and diagonal elements of spatial covariance matrices.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        normalization = self.normalization
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        assert normalization, "Set normalization."

        if type(normalization) is bool:
            # when normalization is True
            normalization = "power"

        if normalization == "power":
            self.normalize_by_power(flooring_fn=flooring_fn)
        else:
            raise NotImplementedError("Normalization {} is not implemented.".format(normalization))

    def normalize_by_power(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Normalize diagonalizers and diagonal elements of spatial covariance matrices by power.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        Diagonalizers are normalized by

        .. math::
            \boldsymbol{q}_{im}
            \leftarrow\frac{\boldsymbol{q}_{im}}{\psi_{im}},

        where

        .. math::
            \psi_{im}
            = \sqrt{\frac{1}{IJ}\sum_{i,j}|\boldsymbol{q}_{im}^{\mathsf{H}}
            \boldsymbol{x}_{ij}|^{2}}.

        For diagonal elements of spatial covariance matrices,

        .. math::
            d_{inm}
            \leftarrow\frac{d_{inm}}{\psi_{im}^{2}}.
        """
        X = self.input
        Q, D = self.diagonalizer, self.spatial
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        QX = Q @ X.transpose(1, 0, 2)
        QX2 = np.mean(np.abs(QX) ** 2, axis=(0, 2))
        psi = np.sqrt(QX2)
        psi = flooring_fn(psi)

        Q = Q / psi[np.newaxis, :, np.newaxis]
        D = D / (psi**2)

        self.diagonalizer, self.spatial = Q, D


class GaussMNMF(MNMF):
    def __init__(
        self,
        n_basis: int,
        n_sources: Optional[int] = None,
        partitioning: bool = False,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        callbacks: Optional[
            Union[Callable[["GaussMNMF"], None], List[Callable[["GaussMNMF"], None]]]
        ] = None,
        normalization: Union[bool, str] = True,
        record_loss: bool = True,
        reference_id: int = 0,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        super().__init__(
            n_basis,
            n_sources=n_sources,
            partitioning=partitioning,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            normalization=normalization,
            record_loss=record_loss,
            reference_id=reference_id,
            rng=rng,
        )

    def __repr__(self) -> str:
        s = "GaussMNMF("
        s += "n_basis={n_basis}"

        if self.n_sources is not None:
            s += ", n_sources={n_sources}"

        if hasattr(self, "n_channels"):
            s += ", n_channels={n_channels}"

        s += ", partitioning={partitioning}"
        s += ", normalization={normalization}"
        s += ", record_loss={record_loss}"
        s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def separate(self, input: np.ndarray) -> np.ndarray:
        """Separate ``input`` using multichannel Wiener filter.

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_sources, n_bins, n_frames).
        """
        n_sources = self.n_sources
        reference_id = self.reference_id

        X = input
        T, V = self.basis, self.activation
        H = self.spatial

        if self.partitioning:
            Lamb = self.reconstruct_nmf(T, V, latent=self.latent)
        else:
            Lamb = self.reconstruct_nmf(T, V)

        R_n = Lamb[:, :, :, np.newaxis, np.newaxis] * H[:, :, np.newaxis, :, :]
        R = np.sum(R_n, axis=0)
        R = to_psd(R, flooring_fn=self.flooring_fn)
        R = np.tile(R, reps=(n_sources, 1, 1, 1, 1))
        W_Hermite = np.linalg.solve(R, R_n)
        W = W_Hermite.transpose(0, 1, 2, 4, 3).conj()
        W_ref = W[:, :, :, reference_id, :]
        W_ref = W_ref.transpose(0, 3, 1, 2)
        Y = np.sum(W_ref * X, axis=1)

        return Y

    def compute_loss(self) -> float:
        r"""Compute loss :math:`\mathcal{L}`.

        Returns:
            Computed loss.
        """
        XX = self.instant_covariance
        T, V = self.basis, self.activation
        H = self.spatial

        if self.partitioning:
            R = self.reconstruct_mnmf(T, V, H, latent=self.latent)
        else:
            R = self.reconstruct_mnmf(T, V, H)

        R = to_psd(R, flooring_fn=self.flooring_fn)
        XXR_inv = np.linalg.solve(R, XX)  # Hermitian transpose of XX @ np.linalg.inv(R)
        trace = np.trace(XXR_inv, axis1=-2, axis2=-1)
        trace = np.real(trace)
        logdet = self.compute_logdet(R)
        loss = np.mean(trace + logdet, axis=-1)
        loss = loss.sum(axis=0)
        loss = loss.item()

        return loss

    def compute_logdet(self, reconstructed: np.ndarray) -> np.ndarray:
        r"""Compute log-determinant.

        Args:
            reconstructed:
                Reconstructed MNMF with shape of (\*, n_channels, n_channels).

        Returns:
            numpy.ndarray of computed log-determinant values.
            The shape is (\*).
        """
        _, logdet = np.linalg.slogdet(reconstructed)

        return logdet

    def update_once(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update MNMF parameters once.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        self.update_basis(flooring_fn=flooring_fn)
        self.update_activation(flooring_fn=flooring_fn)
        self.update_spatial(flooring_fn=flooring_fn)

        if self.normalization:
            # ensure unit trace of spatial property
            # before updates of latent variables in MNMF
            self.normalize(axis1=-2, axis2=-1)

        if self.partitioning:
            self.update_latent(flooring_fn=flooring_fn)

    def update_basis(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update NMF bases by MM algorithm.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        n_sources = self.n_sources
        n_frames = self.n_frames
        na = np.newaxis

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        def _compute_traces(
            target: np.ndarray, reconstructed: np.ndarray, spatial: np.ndarray
        ) -> np.ndarray:
            RXX = np.linalg.solve(reconstructed, target)
            R = np.tile(reconstructed, reps=(n_sources, 1, 1, 1, 1))
            H = np.tile(spatial[:, :, na, :, :], reps=(1, 1, n_frames, 1, 1))
            RH = np.linalg.solve(R, H)

            trace_RXXRH = np.trace(RXX @ RH, axis1=-2, axis2=-1)
            trace_RXXRH = np.real(trace_RXXRH)
            trace_RH = np.trace(RH, axis1=-2, axis2=-1)
            trace_RH = np.real(trace_RH)

            return trace_RXXRH, trace_RH

        XX = self.instant_covariance
        T, V = self.basis, self.activation
        H = self.spatial

        if self.partitioning:
            Z = self.latent
            R = self.reconstruct_mnmf(T, V, H, latent=Z)
            R = to_psd(R, flooring_fn=flooring_fn)

            trace_RXXRH, trace_RH = _compute_traces(XX, R, spatial=H)

            VRXXRH = np.sum(V[na, na, :] * trace_RXXRH[:, :, na], axis=-1)
            VRH = np.sum(V[na, na, :] * trace_RH[:, :, na], axis=-1)

            num = np.sum(Z[:, na, :] * VRXXRH, axis=0)
            denom = np.sum(Z[:, na, :] * VRH, axis=0)
        else:
            R = self.reconstruct_mnmf(T, V, H)
            R = to_psd(R, flooring_fn=flooring_fn)

            trace_RXXRH, trace_RH = _compute_traces(XX, R, spatial=H)

            num = np.sum(V[:, na, :, :] * trace_RXXRH[:, :, na, :], axis=-1)
            denom = np.sum(V[:, na, :, :] * trace_RH[:, :, na, :], axis=-1)

        T = T * np.sqrt(num / denom)
        T = flooring_fn(T)

        self.basis = T

    def update_activation(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update NMF activations by MM algorithm.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        n_sources = self.n_sources
        n_frames = self.n_frames
        na = np.newaxis

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        def _compute_traces(
            target: np.ndarray, reconstructed: np.ndarray, spatial: np.ndarray
        ) -> np.ndarray:
            RXX = np.linalg.solve(reconstructed, target)
            R = np.tile(reconstructed, reps=(n_sources, 1, 1, 1, 1))
            H = np.tile(spatial[:, :, na, :, :], reps=(1, 1, n_frames, 1, 1))
            RH = np.linalg.solve(R, H)

            trace_RXXRH = np.trace(RXX @ RH, axis1=-2, axis2=-1)
            trace_RXXRH = np.real(trace_RXXRH)
            trace_RH = np.trace(RH, axis1=-2, axis2=-1)
            trace_RH = np.real(trace_RH)

            return trace_RXXRH, trace_RH

        XX = self.instant_covariance
        T, V = self.basis, self.activation
        H = self.spatial

        if self.partitioning:
            Z = self.latent
            R = self.reconstruct_mnmf(T, V, H, latent=Z)
            R = to_psd(R, flooring_fn=flooring_fn)

            trace_RXXRH, trace_RH = _compute_traces(XX, R, spatial=H)

            TRXXRH = np.sum(T[na, :, :, na] * trace_RXXRH[:, :, na, :], axis=1)
            TRH = np.sum(T[na, :, :, na] * trace_RH[:, :, na, :], axis=1)

            num = np.sum(Z[:, :, na] * TRXXRH, axis=0)
            denom = np.sum(Z[:, :, na] * TRH, axis=0)
        else:
            R = self.reconstruct_mnmf(T, V, H)
            R = to_psd(R, flooring_fn=flooring_fn)

            trace_RXXRH, trace_RH = _compute_traces(XX, R, spatial=H)

            num = np.sum(T[:, :, :, na] * trace_RXXRH[:, :, na, :], axis=1)
            denom = np.sum(T[:, :, :, na] * trace_RH[:, :, na, :], axis=1)

        V = V * np.sqrt(num / denom)
        V = flooring_fn(V)

        self.activation = V

    def update_spatial(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update spatial properties in NMF by MM algorithm.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        na = np.newaxis
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        XX = self.instant_covariance
        T, V = self.basis, self.activation
        H = self.spatial

        if self.partitioning:
            Z = self.latent
            Lamb = self.reconstruct_nmf(T, V, latent=Z)
        else:
            Lamb = self.reconstruct_nmf(T, V)

        R_n = Lamb[:, :, :, na, na] * H[:, :, na, :, :]
        R = np.sum(R_n, axis=0)
        R = to_psd(R, flooring_fn=flooring_fn)
        R_inverse = np.linalg.inv(R)
        RXXR = R_inverse @ XX @ R_inverse

        P = np.sum(Lamb[:, :, :, na, na] * R_inverse, axis=2)
        Q = np.sum(Lamb[:, :, :, na, na] * RXXR, axis=2)
        HQH = H @ Q @ H

        P = to_psd(P, flooring_fn=flooring_fn)
        HQH = to_psd(HQH, flooring_fn=flooring_fn)

        # geometric mean of P^(-1) and HQH
        H = gmeanmh(P, HQH, type=2)
        H = to_psd(H, flooring_fn=flooring_fn)

        self.spatial = H

    def update_latent(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update latent variables in NMF by MM algorithm.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        n_sources = self.n_sources
        n_frames = self.n_frames
        na = np.newaxis

        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        def _compute_traces(
            target: np.ndarray, reconstructed: np.ndarray, spatial: np.ndarray
        ) -> np.ndarray:
            RXX = np.linalg.solve(reconstructed, target)
            R = np.tile(reconstructed, reps=(n_sources, 1, 1, 1, 1))
            H = np.tile(spatial[:, :, na, :, :], reps=(1, 1, n_frames, 1, 1))
            RH = np.linalg.solve(R, H)

            trace_RXXRH = np.trace(RXX @ RH, axis1=-2, axis2=-1)
            trace_RXXRH = np.real(trace_RXXRH)
            trace_RH = np.trace(RH, axis1=-2, axis2=-1)
            trace_RH = np.real(trace_RH)

            return trace_RXXRH, trace_RH

        XX = self.instant_covariance
        T, V = self.basis, self.activation
        H, Z = self.spatial, self.latent

        R = self.reconstruct_mnmf(T, V, H, latent=Z)
        R = to_psd(R, flooring_fn=flooring_fn)

        trace_RXXRH, trace_RH = _compute_traces(XX, R, spatial=H)

        VRXXRH = np.sum(V[na, na, :] * trace_RXXRH[:, :, na], axis=-1)
        VRH = np.sum(V[na, na, :] * trace_RH[:, :, na], axis=-1)

        num = np.sum(T * VRXXRH, axis=1)
        denom = np.sum(T * VRH, axis=1)

        Z = Z * np.sqrt(num / denom)
        Z = Z / Z.sum(axis=0)

        self.latent = Z


class FastGaussMNMF(FastMNMFBase):
    r"""Fast multichannel nonnegative matrix factorization on Gaussian distribution \
    (Fast Gauss-MNMF).

    Args:
        n_basis (int):
            Number of NMF bases.
        n_sources (int, optional):
            Number of sources to be separated.
            If ``None`` is given, ``n_sources`` is determined by number of channels
            in input spectrogram. Default: ``None``.
        diagonalizer_algorithm (str):
            Algorithm for diagonalizers. Choose ``IP``, ``IP1``, or ``IP2``.
            Default: ``IP``.
        partitioning (bool):
            Whether to use partioning function. Default: ``False``.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel in multichannel Wiener filter. Default: ``0``.
        rng (numpy.random.Generator, optioinal):
            Random number generator. This is mainly used to randomly initialize PSDTF.
            If ``None`` is given, ``np.random.default_rng()`` is used.
            Default: ``None``.
    """

    def __init__(
        self,
        n_basis: int,
        n_sources: Optional[int] = None,
        diagonalizer_algorithm: str = "IP",
        partitioning: bool = False,
        flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
            max_flooring, eps=EPS
        ),
        pair_selector: Optional[Callable[[int], Iterable[Tuple[int, int]]]] = None,
        callbacks: Optional[
            Union[Callable[["FastGaussMNMF"], None], List[Callable[["FastGaussMNMF"], None]]]
        ] = None,
        normalization: bool = True,
        record_loss: bool = True,
        reference_id: int = 0,
        rng: Optional[np.random.Generator] = None,
    ) -> None:
        super().__init__(
            n_basis,
            n_sources=n_sources,
            partitioning=partitioning,
            flooring_fn=flooring_fn,
            callbacks=callbacks,
            normalization=normalization,
            record_loss=record_loss,
            reference_id=reference_id,
            rng=rng,
        )

        assert diagonalizer_algorithm in diagonalizer_algorithms, "Not support {}.".format(
            diagonalizer_algorithm
        )
        assert not partitioning, "partitioning function is not supported."

        self.diagonalizer_algorithm = diagonalizer_algorithm

        if pair_selector is None:
            if diagonalizer_algorithm == "IP2":
                self.pair_selector = sequential_pair_selector
        else:
            self.pair_selector = pair_selector

    def __repr__(self) -> str:
        s = "FastGaussMNMF("
        s += "n_basis={n_basis}"

        if self.n_sources is not None:
            s += ", n_sources={n_sources}"

        if hasattr(self, "n_channels"):
            s += ", n_channels={n_channels}"

        s += ", diagonalizer_algorithm={diagonalizer_algorithm}"
        s += ", partitioning={partitioning}"
        s += ", record_loss={record_loss}"
        s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def separate(self, input: np.ndarray) -> np.ndarray:
        """Separate ``input`` using multichannel Wiener filter.

        Args:
            input (numpy.ndarray):
                The mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_sources, n_bins, n_frames).
        """
        na = np.newaxis
        n_sources = self.n_sources
        reference_id = self.reference_id

        X = input
        T, V = self.basis, self.activation
        Q, D = self.diagonalizer, self.spatial

        if self.partitioning:
            Lamb = self.reconstruct_nmf(T, V, latent=self.latent)
        else:
            Lamb = self.reconstruct_nmf(T, V)

        D = D.transpose(1, 0, 2)

        Q_inverse = np.linalg.inv(Q)
        Q_inverse_Hermite = Q_inverse.transpose(0, 2, 1).conj()
        QQ_Hermite = Q_inverse[:, :, :, na] * Q_inverse_Hermite[:, na, :, :]

        LambD = Lamb[:, :, :, na] * D[:, :, na, :]

        R_n = np.sum(LambD[:, :, :, na, :, na] * QQ_Hermite[:, na, :, :, :], axis=4)
        R = np.sum(R_n, axis=0)
        R = to_psd(R, flooring_fn=self.flooring_fn)
        R = np.tile(R, reps=(n_sources, 1, 1, 1, 1))
        W_Hermite = np.linalg.solve(R, R_n)
        W = W_Hermite.transpose(0, 1, 2, 4, 3).conj()
        W_ref = W[:, :, :, reference_id, :]
        W_ref = W_ref.transpose(0, 3, 1, 2)
        Y = np.sum(W_ref * X, axis=1)

        return Y

    def compute_loss(self) -> float:
        r"""Compute loss :math:`\mathcal{L}`.

        :math:`\mathcal{L}` is defined as follows:

        .. math::
            \mathcal{L}
            &:=-\frac{1}{J}\sum_{i,j}\left\{
            \mathrm{tr}\left(
            \boldsymbol{x}_{ij}\boldsymbol{x}_{ij}^{\mathsf{H}}\boldsymbol{R}_{ij}^{-1}
            \right)
            - \log\det\boldsymbol{R}_{ij}
            \right\} \\
            &:=\frac{1}{J}\sum_{i,j,m}\left\{
            \frac{|\boldsymbol{q}_{im}^{\mathsf{H}}\boldsymbol{x}_{ij}|^{2}}
            {\sum_{n}\lambda_{ijn}d_{inm}}
            + \log\sum_{n}\lambda_{ijn}d_{inm}\right\}
            - 2\sum_{i}\log|\det\boldsymbol{Q}_{i}|.

        Returns:
            Computed loss.
        """
        X = self.input
        T, V = self.basis, self.activation
        Q, D = self.diagonalizer, self.spatial
        na = np.newaxis

        if self.partitioning:
            Lamb = self.reconstruct_nmf(T, V, latent=self.latent)
        else:
            Lamb = self.reconstruct_nmf(T, V)

        D = D.transpose(1, 0, 2)
        LambD = np.sum(Lamb[:, :, na, :] * D[:, :, :, na], axis=0)
        QX = Q @ X.transpose(1, 0, 2)
        QX2 = np.abs(QX) ** 2
        logdetQ = self.compute_logdet(Q)
        loss = np.sum(QX2 / LambD + np.log(LambD), axis=1)
        loss = np.mean(loss, axis=-1) - 2 * logdetQ
        loss = loss.sum(axis=0)
        loss = loss.item()

        return loss

    def compute_logdet(self, diagonalizer: np.ndarray) -> np.ndarray:
        r"""Compute log-determinant.

        Args:
            reconstructed:
                Diagonalizer with shape of (\*, n_channels, n_channels).

        Returns:
            numpy.ndarray of computed log-determinant values.
            The shape is (\*).
        """
        _, logdet = np.linalg.slogdet(diagonalizer)

        return logdet

    def update_once(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update MNMF parameters, diagonalizers, and diagonal elements of \
        spatial covariance matrices once.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        self.update_basis(flooring_fn=flooring_fn)
        self.update_activation(flooring_fn=flooring_fn)
        self.update_diagonalizer(flooring_fn=flooring_fn)
        self.update_spatial()

        if self.normalization:
            self.normalize(flooring_fn=flooring_fn)

    def update_basis(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update NMF bases by MM algorithm.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        Update :math:`t_{ikn}` as follows:

        .. math::
            t_{ikn}
            \leftarrow\left[
            \frac{\displaystyle\sum_{j,m}\frac{|\boldsymbol{q}_{im}^{\mathsf{H}}\boldsymbol{x}_{ij}|^{2}d_{inm}v_{kjn}}
            {\left(\sum_{k',n'}t_{ik'n'}v_{k'jn'}d_{in'm}\right)^{2}}}
            {\displaystyle\sum_{j,m}\dfrac{d_{inm}v_{kjn}}{\sum_{k',n'}t_{ik'n'}v_{k'jn'}d_{in'm}}}
            \right]^{\frac{1}{2}}t_{ikn}.
        """
        assert not self.partitioning, "partitioning function is not supported."

        na = np.newaxis
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        X = self.input
        T, V = self.basis, self.activation
        Q, D = self.diagonalizer, self.spatial

        if self.partitioning:
            Lamb = self.reconstruct_nmf(T, V, latent=self.latent)
        else:
            Lamb = self.reconstruct_nmf(T, V)

        D = D.transpose(1, 0, 2)
        LambD = Lamb[:, :, :, na] * D[:, :, na, :]
        LambD = np.sum(LambD, axis=0)
        QX = Q @ X.transpose(1, 0, 2)
        QX = np.abs(QX)
        QX = QX.transpose(0, 2, 1)
        QXLambD = (QX / LambD) ** 2
        DQXLambD = np.sum(D[:, :, na, :] * QXLambD, axis=-1)
        DLambD = np.sum(D[:, :, na, :] / LambD, axis=-1)

        num = np.sum(V[:, na, :] * DQXLambD[:, :, na], axis=-1)
        denom = np.sum(V[:, na, :] * DLambD[:, :, na], axis=-1)

        T = T * np.sqrt(num / denom)
        T = flooring_fn(T)

        self.basis = T

    def update_activation(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update NMF activations by MM algorithm.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        Update :math:`v_{kjn}` as follows:

        .. math::
            v_{kjn}
            \leftarrow\left[
            \frac{\displaystyle\sum_{i,m}\frac{|\boldsymbol{q}_{im}^{\mathsf{H}}\boldsymbol{x}_{ij}|^{2}d_{inm}t_{ikn}}
            {\left(\sum_{k',n'}t_{ik'n'}v_{k'jn'}d_{in'm}\right)^{2}}}
            {\displaystyle\sum_{i,m}\dfrac{d_{inm}t_{ikn}}{\sum_{k',n'}t_{ik'n'}v_{k'jn'}d_{in'm}}}
            \right]^{\frac{1}{2}}v_{kjn}.
        """
        assert not self.partitioning, "partitioning function is not supported."

        na = np.newaxis
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        X = self.input
        T, V = self.basis, self.activation
        Q, D = self.diagonalizer, self.spatial

        if self.partitioning:
            Lamb = self.reconstruct_nmf(T, V, latent=self.latent)
        else:
            Lamb = self.reconstruct_nmf(T, V)

        D = D.transpose(1, 0, 2)
        LambD = Lamb[:, :, :, na] * D[:, :, na, :]
        LambD = np.sum(LambD, axis=0)
        QX = Q @ X.transpose(1, 0, 2)
        QX = np.abs(QX)
        QX = QX.transpose(0, 2, 1)
        QXLambD = (QX / LambD) ** 2
        DQXLambD = np.sum(D[:, :, na, :] * QXLambD, axis=-1)
        DLambD = np.sum(D[:, :, na, :] / LambD, axis=-1)

        num = np.sum(T[:, :, :, na] * DQXLambD[:, :, na, :], axis=1)
        denom = np.sum(T[:, :, :, na] * DLambD[:, :, na, :], axis=1)

        V = V * np.sqrt(num / denom)
        V = flooring_fn(V)

        self.activation = V

    def update_diagonalizer(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        """Update diagonalizer.

        - If ``diagonalizer_algorithm`` is ``IP`` or ``IP1``, \
            ``update_diagonalizer_model_ip1`` is called.
        - If ``diagonalizer_algorithm`` is ``IP2``, \
            ``update_diagonalizer_model_ip2`` is called.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        """
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        if self.diagonalizer_algorithm in ["IP", "IP1"]:
            self.update_diagonalizer_ip1(flooring_fn=flooring_fn)
        elif self.diagonalizer_algorithm in ["IP2"]:
            self.update_diagonalizer_ip2(flooring_fn=flooring_fn)
        else:
            raise NotImplementedError("Not support {}.".format(self.diagonalizer_algorithm))

    def update_diagonalizer_ip1(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update diagonalizer once using iterative projection.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        Diagonalizers are updated sequentially for :math:`m=1,\ldots,M` as follows:

        .. math::
            \boldsymbol{q}_{im}
            &\leftarrow\left(\boldsymbol{Q}_{im}^{\mathsf{H}}\boldsymbol{U}_{im}\right)^{-1} \
            \boldsymbol{e}_{m}, \\
            \boldsymbol{q}_{im}
            &\leftarrow\frac{\boldsymbol{q}_{im}}
            {\sqrt{\boldsymbol{q}_{im}^{\mathsf{H}}\boldsymbol{U}_{im}\boldsymbol{q}_{im}}},

        where

        .. math::
            \boldsymbol{U}_{im}
            = \frac{1}{J}\sum_{j}
            \frac{\boldsymbol{x}_{ij}\boldsymbol{x}_{ij}^{\mathsf{H}}}
            {\sum_{n}\left(\sum_{k}z_{nk}t_{ik}v_{kj}\right)d_{inm}}

        if ``partitioning=True``, otherwise

        .. math::
            \boldsymbol{U}_{im}
            = \frac{1}{J}\sum_{j}
            \frac{\boldsymbol{x}_{ij}\boldsymbol{x}_{ij}^{\mathsf{H}}}
            {\sum_{n}\left(\sum_{k}t_{ikn}v_{kjn}\right)d_{inm}}.
        """
        assert not self.partitioning, "partitioning function is not supported."

        na = np.newaxis
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        X = self.input
        T, V = self.basis, self.activation
        Q, D = self.diagonalizer, self.spatial

        if self.partitioning:
            Lamb = self.reconstruct_nmf(T, V, latent=self.latent)
        else:
            Lamb = self.reconstruct_nmf(T, V)

        XX_Hermite = X[:, na, :, :] * X[na, :, :, :].conj()
        XX_Hermite = XX_Hermite.transpose(2, 0, 1, 3)

        Lamb = Lamb.transpose(1, 0, 2)
        LambD = np.sum(Lamb[:, :, na, :] * D[:, :, :, na], axis=1)
        varphi = 1 / LambD

        varphi_XX = varphi[:, :, na, na, :] * XX_Hermite[:, na, :, :, :]
        U = np.mean(varphi_XX, axis=-1)

        self.diagonalizer = update_by_ip1(Q, U, flooring_fn=flooring_fn)

    def update_diagonalizer_ip2(
        self,
        flooring_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "self",
    ) -> None:
        r"""Update diagonalizer once using pairwise iterative projection.

        Args:
            flooring_fn (callable or str, optional):
                A flooring function for numerical stability.
                This function is expected to return the same shape tensor as the input.
                If you explicitly set ``flooring_fn=None``,
                the identity function (``lambda x: x``) is used.
                If ``self`` is given as str, ``self.flooring_fn`` is used.
                Default: ``self``.

        For :math:`m_{1}` and :math:`m_{2}` (:math:`m_{1}\neq m_{2}`),
        compute weighted covariance matrix as follows:

        .. math::
            \boldsymbol{U}_{im}
            = \frac{1}{J}\sum_{j}
            \frac{\boldsymbol{x}_{ij}\boldsymbol{x}_{ij}^{\mathsf{H}}}{\sum_{n}\lambda_{ijn}d_{inm}},

        :math:`\lambda_{ijn}` is computed by

        .. math::
            \lambda_{ijn}=\sum_{k}z_{nk}t_{ik}v_{kj}

        if ``partitioning=True``.
        Otherwise,

        .. math::
            \lambda_{ijn}=\sum_{k}t_{ikn}v_{kjn}.

        Using :math:`\boldsymbol{U}_{im_{1}}` and
        :math:`\boldsymbol{U}_{im_{2}}`, we compute generalized eigenvectors.

        .. math::
            \left({\boldsymbol{P}_{im_{1}}^{(m_{1},m_{2})}}^{\mathsf{H}}\boldsymbol{U}_{im_{1}}
            \boldsymbol{P}_{im_{1}}^{(m_{1},m_{2})}\right)\boldsymbol{h}_{i}
            = \mu_{i}
            \left({\boldsymbol{P}_{im_{2}}^{(m_{1},m_{2})}}^{\mathsf{H}}\boldsymbol{U}_{im_{2}}
            \boldsymbol{P}_{im_{2}}^{(m_{1},m_{2})}\right)\boldsymbol{h}_{i},

        where

        .. math::
            \boldsymbol{P}_{im_{1}}^{(m_{1},m_{2})}
            &= (\boldsymbol{Q}_{i}\boldsymbol{U}_{im_{1}})^{-1}
            (
            \begin{array}{cc}
                \boldsymbol{e}_{m_{1}} & \boldsymbol{e}_{m_{2}}
            \end{array}
            ), \\
            \boldsymbol{P}_{im_{2}}^{(m_{1},m_{2})}
            &= (\boldsymbol{Q}_{i}\boldsymbol{U}_{im_{2}})^{-1}
            (
            \begin{array}{cc}
                \boldsymbol{e}_{m_{1}} & \boldsymbol{e}_{m_{2}}
            \end{array}
            ).

        After that, we standardize two eigenvectors :math:`\boldsymbol{h}_{im_{1}}`
        and :math:`\boldsymbol{h}_{im_{2}}`.

        .. math::
            \boldsymbol{h}_{im_{1}}
            &\leftarrow\frac{\boldsymbol{h}_{im_{1}}}
            {\sqrt{\boldsymbol{h}_{im_{1}}^{\mathsf{H}}
            \left({\boldsymbol{P}_{im_{1}}^{(m_{1},m_{2})}}^{\mathsf{H}}\boldsymbol{U}_{im_{1}}
            \boldsymbol{P}_{im_{1}}^{(m_{1},m_{2})}\right)
            \boldsymbol{h}_{im_{1}}}}, \\
            \boldsymbol{h}_{im_{2}}
            &\leftarrow\frac{\boldsymbol{h}_{im_{2}}}
            {\sqrt{\boldsymbol{h}_{im_{2}}^{\mathsf{H}}
            \left({\boldsymbol{P}_{im_{2}}^{(m_{1},m_{2})}}^{\mathsf{H}}\boldsymbol{U}_{im_{2}}
            \boldsymbol{P}_{im_{2}}^{(m_{1},m_{2})}\right)
            \boldsymbol{h}_{im_{2}}}}.

        Then, update :math:`\boldsymbol{q}_{im_{1}}` and :math:`\boldsymbol{q}_{im_{2}}`
        simultaneously.

        .. math::
            \boldsymbol{q}_{im_{1}}
            &\leftarrow \boldsymbol{P}_{im_{1}}^{(m_{1},m_{2})}\boldsymbol{h}_{im_{1}} \\
            \boldsymbol{q}_{im_{2}}
            &\leftarrow \boldsymbol{P}_{im_{2}}^{(m_{1},m_{2})}\boldsymbol{h}_{im_{2}}

        At each iteration, we update pairs of :math:`m_{1}` and :math:`m_{2}`
        for :math:`m_{1}\neq m_{2}`.
        """
        assert not self.partitioning, "partitioning function is not supported."

        na = np.newaxis
        flooring_fn = choose_flooring_fn(flooring_fn, method=self)

        X = self.input
        T, V = self.basis, self.activation
        Q, D = self.diagonalizer, self.spatial

        if self.partitioning:
            Lamb = self.reconstruct_nmf(T, V, latent=self.latent)
        else:
            Lamb = self.reconstruct_nmf(T, V)

        XX_Hermite = X[:, na, :, :] * X[na, :, :, :].conj()
        XX_Hermite = XX_Hermite.transpose(2, 0, 1, 3)

        Lamb = Lamb.transpose(1, 0, 2)
        LambD = np.sum(Lamb[:, :, na, :] * D[:, :, :, na], axis=1)
        varphi = 1 / LambD

        varphi_XX = varphi[:, :, na, na, :] * XX_Hermite[:, na, :, :, :]
        U = np.mean(varphi_XX, axis=-1)

        self.diagonalizer = update_by_ip2(
            Q, U, flooring_fn=flooring_fn, pair_selector=self.pair_selector
        )

    def update_spatial(self) -> None:
        r"""Update diagonal elements of spatial covariance matrix by MM algorithm.

        Update :math:`d_{inm}` as follows:

        .. math::
            d_{inm}\leftarrow\left[
            \dfrac{\displaystyle\sum_{j}\frac{\lambda_{ijn}|\boldsymbol{q}_{im}^{\mathsf{H}}\boldsymbol{x}_{ij}|^{2}}
            {\left(\sum_{n'}\lambda_{ijn'}d_{in'm}\right)^{2}}}
            {\displaystyle\sum_{j}\frac{\lambda_{ijn}}{\sum_{n'}\lambda_{ijn'}d_{in'm}}}
            \right]^{\frac{1}{2}}d_{inm}.
        """
        assert not self.partitioning, "partitioning function is not supported."

        na = np.newaxis

        X = self.input
        T, V = self.basis, self.activation
        Q, D = self.diagonalizer, self.spatial

        if self.partitioning:
            Lamb = self.reconstruct_nmf(T, V, latent=self.latent)
        else:
            Lamb = self.reconstruct_nmf(T, V)

        QX = Q @ X.transpose(1, 0, 2)
        QX = np.abs(QX)
        QX2 = QX**2

        Lamb = Lamb.transpose(1, 0, 2)
        LambD = np.sum(Lamb[:, :, na, :] * D[:, :, :, na], axis=1)
        LambD2 = LambD**2
        Lamb_LambD2 = Lamb[:, :, na] / LambD2[:, na, :]
        num = np.sum(Lamb_LambD2 * QX2[:, na, :, :], axis=-1)

        Lamb_LambD = Lamb[:, :, na] / LambD[:, na, :]
        denom = np.sum(Lamb_LambD, axis=-1)

        D = np.sqrt(num / denom) * D

        self.spatial = D



from typing import Optional

import numpy as np
from numpy.linalg import LinAlgError

from .cubic import cbrt


def solve_cubic(
    A: np.ndarray,
    B: np.ndarray,
    C: np.ndarray,
    D: Optional[np.ndarray] = None,
    all: bool = True,
) -> np.ndarray:
    r"""Find roots of cubic equations.

    Args:
        A (numpy.ndarray): Coefficients of cubic or quadratic terms.
        B (numpy.ndarray): Coefficients of quadratic or linear terms.
        C (numpy.ndarray): Coefficients of linear or constant terms.
        D (numpy.ndarray, optional): Constant terms.
        all (bool): If ``all=True``, returns all roots. Otherwise, returns one of them.
            Default: ``True``.

    Returns:
        numpy.ndarray: All roots of cuadratic equations of shape (3, \*) if ``all=True``.
            Otherwise, (\*).

    This function solves the following equations if ``D`` is given:

    .. math::

        Ax^{3} + Bx^{2} + Cx + D = 0.

    If ``D`` is not given, solves

    .. math::

        x^{3} + Ax^{2} + Bx + C = 0.

    """
    if D is None:
        P = -(A**2) / 3 + B
        Q = (2 * A**3) / 27 - (A * B) / 3 + C

        X = _find_cubic_roots(P, Q)
        x = X - A / 3

        return x if all else x[0]
    else:
        if np.any(A == 0):
            raise LinAlgError("Coefficients include zero.")

        return solve_cubic(B / A, C / A, D / A, all=all)


def _find_cubic_roots(P: np.ndarray, Q: np.ndarray) -> np.ndarray:
    r"""Find roots of the following cubic equations:

    .. math::

        x^{3} + px + q = 0


    Args:
        P (np.ndarray): Coefficients of cubic equation.
        Q (np.ndarray): Coefficients of cubic equation.

    Returns:
        numpy.ndarray of the three roots.
        The shape is (3, \*).

    """
    P = P.astype(np.complex128)
    Q = Q.astype(np.complex128)
    omega = (-1 + 1j * np.sqrt(3)) / 2
    omega_conj = (-1 - 1j * np.sqrt(3)) / 2

    discriminant = (Q / 2) ** 2 + (P / 3) ** 3

    U = cbrt(-Q / 2 + np.sqrt(discriminant))
    # U = 0, when P = 0.
    is_singular = P == 0
    U = np.where(is_singular, 1, U)
    V = -P / (3 * U)

    X1 = U + V
    X1 = np.where(is_singular, cbrt(-Q), X1)
    X2 = U * omega + V * omega_conj
    X2 = np.where(is_singular, X1 * omega, X2)
    X3 = U * omega_conj + V * omega
    X3 = np.where(is_singular, X1 * omega_conj, X3)

    return np.stack([X1, X2, X3], axis=0)



from typing import Optional

import numpy as np


def minimal_distortion_principle(
    estimated: np.ndarray,
    reference: Optional[np.ndarray] = None,
    reference_id: Optional[int] = 0,
) -> np.ndarray:
    r"""Minimal distortion principle to restore scale ambiguity.

    The implementation is based on [#matsuoka2002minimal]_.

    Args:
        estimated (numpy.ndarray):
            Estimated spectrograms with shape of (n_channels, n_bins, n_frames).
        reference (numpy.ndarray, optional):
            Reference spectrogram with shape of (n_sources, n_bins, n_frames).
        reference_id (int, optional):
            Reference microphone index. Default: ``0``.

    Returns:
        numpy.ndarray of rescaled estimated spectrograms or demixing filters.

    .. [#matsuoka2002minimal]
        N. Murata, S. Ikeda, and A. Ziehe,
        "Minimal distortion principle for blind source separation,"
        in *Proc. ICA*, 2001, pp. 722-727.
    """
    Y = estimated
    X_conj = reference.conj()

    if reference_id is None:
        num = np.sum(Y * X_conj[:, np.newaxis, :, :], axis=-1, keepdims=True)
    else:
        num = np.sum(Y * X_conj[reference_id], axis=-1, keepdims=True)

    denom = np.sum(np.abs(Y) ** 2, axis=-1, keepdims=True)
    Z = num / denom
    output_scaled = Z.conj() * Y

    return output_scaled



from .flooring import add_flooring, identity, max_flooring
from .logsumexp import logsumexp
from .psd import to_psd
from .softmax import softmax

__all__ = ["add_flooring", "max_flooring", "identity", "to_psd", "logsumexp", "softmax"]



from typing import Tuple

import numpy as np

from .mird import download as download_mird
from .sisec2010 import download as download_sisec2010

__all__ = ["download_sample_speech_data"]

sisec2010_tags = ["dev1_female3", "dev1_female4"]


def download_sample_speech_data(
    sisec2010_root: str = ".data/SiSEC2010",
    mird_root: str = ".data/MIRD",
    n_sources: int = 3,
    sisec2010_tag: str = "dev1_female3",
    max_duration: float = 10,
    reverb_duration: float = 0.16,
    conv: bool = True,
) -> Tuple[np.ndarray, int]:
    r"""Download sample speech data to test sepration methods.

    This function returns source images of sample speech data.

    Args:
        sisec2010_root (str):
            Path to save SiSEC2010 dataset. Default: ".data/SiSEC2010".
        mird_root (str):
            Path to save MIRD dataset. Default: ".data/MIRD".
        n_sources (int):
            Number of sources included in sample data.
        sisec2010_tag (str):
            Tag of SiSEC 2010 data.
            Choose ``dev1_female3`` or ``dev1_female4``.
            Default: ``dev1_female3``.
        max_duration (float):
            Maximum duration. Default: ``160000``.
        reverb_duration (float):
            Duration of reverberation in MIRD.
            Choose ``0.16``, ``0.36``, ``0.61``. Default: ``0.16``.
        conv (bool):
            Convolutive mixture or not. Defalt: ``True``.

    Returns:
        Tuple of source images and sampling rate.
        The source images is numpy.ndarry with shape of (n_channels, n_sources, n_samples).
    """
    assert sisec2010_tag in sisec2010_tags, "Choose sisec2010_tag from {}".format(sisec2010_tags)
    sample_rate = 16000  # Only 16khz is supported.
    max_samples = int(sample_rate * max_duration)

    sisec2010_npz_path = download_sisec2010(
        root=sisec2010_root, n_sources=n_sources, tag=sisec2010_tag
    )
    sisec2010_npz = np.load(sisec2010_npz_path)

    assert sample_rate == sisec2010_npz["sample_rate"].item(), "Invalid sampling rate is detected."

    if conv:
        mird_npz_path = download_mird(
            root=mird_root, n_sources=n_sources, reverb_duration=reverb_duration
        )
        mird_npz = np.load(mird_npz_path)

        assert sample_rate == mird_npz["sample_rate"].item(), "Invalid sampling rate is detected."

        waveform_src_img = []

        for src_idx in range(n_sources):
            key = "src_{}".format(src_idx + 1)
            waveform_src = sisec2010_npz[key][:max_samples]
            n_samples = len(waveform_src)
            _waveform_src_img = []

            for waveform_rir in mird_npz[key]:
                waveform_conv = np.convolve(waveform_src, waveform_rir)[:n_samples]
                _waveform_src_img.append(waveform_conv)

            _waveform_src_img = np.stack(_waveform_src_img, axis=0)  # (n_channels, n_samples)
            waveform_src_img.append(_waveform_src_img)

        waveform_src_img = np.stack(waveform_src_img, axis=1)  # (n_channels, n_sources, n_samples)
    else:
        waveform_src_img = []

        rng = np.random.default_rng(seed=42)
        mixing = rng.standard_normal((n_sources, n_sources))

        for src_idx in range(n_sources):
            key = "src_{}".format(src_idx + 1)
            _mixing = mixing[:, src_idx]
            waveform_src = sisec2010_npz[key][:max_samples]
            _waveform_src_img = _mixing[:, np.newaxis] * waveform_src
            waveform_src_img.append(_waveform_src_img)

        waveform_src_img = np.stack(waveform_src_img, axis=1)  # (n_channels, n_sources, n_samples)

    return waveform_src_img, sample_rate



import numpy as np


def logsumexp(X: np.ndarray, axis: int = None, keepdims: bool = False) -> np.ndarray:
    r"""Compute log-sum-exp values.

    Args:
        X (np.ndarray):
            Elements to compute log-sum-exp.
        axis (int or tuple[int], optional):
            Axis or axes over which the sum is performed.
            Default: ``None``.
        keepdims (bool):
            If ``True`` is given, ``axis`` dimension(s) is reduced.
            Default: ``False``.

    Returns:
        np.ndarray of log-sum-exp values.

    Examples:

        .. code-block:: python

            >>> import numpy as np

            >>> X = np.array([[1, 2, 3], [4, 5, 6]])
            >>> logsumexp(X, axis=0)
            array([4.04858735, 5.04858735, 6.04858735])
            >>> logsumexp(X, axis=1)
            array([3.40760596, 6.40760596])
    """
    vmax = np.max(X, axis=axis, keepdims=True)
    exp = np.exp(X - vmax)
    sum_exp = exp.sum(axis=axis, keepdims=True)
    v = np.log(sum_exp) + vmax

    if not keepdims:
        v = np.squeeze(v, axis=axis)

    return v



import warnings

import numpy as np

EPS = 1e-10


def identity(input: np.ndarray) -> np.ndarray:
    r"""Identity function."""
    warnings.warn("Use ssspy.special.identity instead.", FutureWarning)

    return input


def max_flooring(input: np.ndarray, eps: float = EPS) -> np.ndarray:
    r"""Max flooring operation."""
    warnings.warn("Use ssspy.special.max_flooring instead.", FutureWarning)

    return np.maximum(input, eps)


def add_flooring(input: np.ndarray, eps: float = EPS) -> np.ndarray:
    r"""Add flooring operation."""
    warnings.warn("Use ssspy.special.add_flooring instead.", FutureWarning)

    return input + eps



import functools
import warnings
from typing import Callable, Optional, Union

import numpy as np

from ..special.flooring import identity, max_flooring
from .cubic import cbrt

EPS = 1e-10


def lqpqm2(
    H: np.ndarray,
    v: np.ndarray,
    z: np.ndarray,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
    singular_fn: Optional[Union[str, Callable[[np.ndarray], np.ndarray]]] = "flooring",
    max_iter: int = 10,
) -> None:
    r"""Solve of log-quadratically penelized quadratic minimization (type 2).

    .. math::

        \check{\boldsymbol{q}}_{in}
        = \min_{\check{\boldsymbol{q}}_{in}}
        ~~\check{\boldsymbol{q}}_{in}^{\mathsf{H}}\check{\boldsymbol{q}}_{in}
        - \log\left((\check{\boldsymbol{q}}_{in}+\boldsymbol{v}_{in})^{\mathsf{H}}
        \boldsymbol{H}_{in}(\check{\boldsymbol{q}}_{in}+\boldsymbol{v}_{in})
        + z_{in}
        \right)

    Args:
        H (numpy.ndarray): Positive semidefinite matrices of shape
            (n_bins, n_sources - 1, n_sources - 1).
        v (numpy.ndarray): Linear terms in LQPQM of shape (n_bins, n_sources - 1).
        z (numpy.ndarray): Constant terms in LQPQM of shape (n_bins,).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        singular_fn (callable, optional):
            A flooring function to return singular condition.
            This function is expected to return the same shape bool tensor as the input.
            If ``singular_fn=None``, ``lambda x: x == 0`` is used.
            Default: ``flooring``.
        max_iter (int):
            Maximum number of Newton-Raphson method. Default: ``10``.

    Returns:
        np.ndarray: Solutions of LQPQM type-2 of shape (n_bins, n_sources - 1).

    """
    if flooring_fn is None:
        flooring_fn = identity

    if singular_fn is None:

        def _is_zero(x: np.ndarray) -> np.ndarray:
            return x == 0

        singular_fn = _is_zero
    elif singular_fn == "flooring":

        def _is_lower_than_floor(x: np.ndarray) -> np.ndarray:
            return x < flooring_fn(0)

        singular_fn = _is_lower_than_floor
    else:
        assert callable(singular_fn), "singular_fn should be callable."

    phi, sigma = np.linalg.eigh(H)
    norm = np.linalg.norm(v, axis=-1)
    is_singular = singular_fn(norm)

    # when v = 0
    phi_singular = phi[is_singular]
    sigma_singular = sigma[is_singular]
    z_singular = z[is_singular]

    phi_max_singular = phi_singular[:, -1]
    sigma_max_singular = sigma_singular[:, -1]
    lamb_singular = np.maximum(z_singular, phi_max_singular)
    scale = (lamb_singular - z_singular) / phi_max_singular
    scale = np.maximum(scale, 0)
    scale = np.sqrt(scale)
    y_singular = scale[..., np.newaxis] * sigma_max_singular

    # when v != 0
    phi_non_singular = phi[~is_singular]
    sigma_non_singular = sigma[~is_singular]
    v_non_singular = v[~is_singular]
    z_non_singular = z[~is_singular]

    v_tilde_non_singular = np.sum(
        sigma_non_singular.conj() * v_non_singular[:, :, np.newaxis], axis=-2
    )
    lamb_non_singular = solve_equation(
        phi_non_singular,
        v_tilde_non_singular,
        z_non_singular,
        flooring_fn=flooring_fn,
        max_iter=max_iter,
        normalization=True,
    )

    num = phi_non_singular * v_tilde_non_singular
    denom = lamb_non_singular[..., np.newaxis] - phi_non_singular
    v_nonsingular = num / denom
    y_non_singular = np.sum(sigma_non_singular * v_nonsingular[:, np.newaxis, :], axis=-1)

    y = np.zeros_like(v)
    y[is_singular] = y_singular
    y[~is_singular] = y_non_singular

    return y


def solve_equation(
    phi: np.ndarray,
    v: np.ndarray,
    z: np.ndarray,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
    max_iter: int = 10,
    normalization: bool = True,
):
    r"""Find largest root of :math:`f(\lambda_{in})`, where

    .. math::

        f(\lambda_{in})
        = \lambda_{in}^{2}\sum_{n'}
        \frac{\phi_{inn'}|\tilde{v}_{inn'}|^{2}}{(\lambda_{in}-\phi_{inn'})^{2}}
        - \lambda_{in} + z_{in}

    Args:
        phi (numpy.ndarray): Eigen values defined in LQPQM of shape (n_bins, n_sources).
        v (numpy.ndarray): Linear term defined in LQPQM of shape (n_bins, n_sources).
        z (numpy.ndarray): Constant term defined in LQPQM of shape (n_bins,).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        max_iter (int): Maximum iteration of Newton-Raphson method. Default: ``10``.
        normalization (bool): If ``True``, coefficients are normalized by ``phi_max``.

    Returns:
        numpy.ndarray of largest root of :math:`f(\lambda_{in})`.
        The shape is (n_bins,).

    """
    if flooring_fn is None:
        flooring_fn = identity

    n_bins, n_sources = phi.shape

    non_zero_mask = phi * np.abs(v) ** 2 >= flooring_fn(0)
    phi = non_zero_mask * phi
    v = non_zero_mask * v

    max_index = np.argmax(phi, axis=-1) + np.arange(0, n_bins * n_sources, n_sources)
    phi_flatten = phi.flatten()
    v_flatten = v.flatten()
    phi_max = phi_flatten[max_index]
    v_max = v_flatten[max_index]
    phi_max = flooring_fn(phi_max)

    if normalization:
        phi_max_original = phi_max
        phi = phi / phi_max[:, np.newaxis]
        v = v / phi_max[:, np.newaxis]
        v_max = v_max / phi_max
        z = z / phi_max
        phi_max = phi_max / phi_max  # i.e. phi_max = 1
    else:
        phi_max_original = None

    # Find largest root of cubic polynomial for initialization
    A = -(phi_max * np.abs(v_max) ** 2 + 2 * phi_max + z)
    B = (phi_max + 2 * z) * phi_max
    C = -(phi_max**2) * z
    lamb = _find_largest_root(A, B, C)

    is_valid = lamb > phi_max
    lamb[~is_valid] = phi_max[~is_valid] + flooring_fn(0)
    lamb = np.maximum(lamb, z)

    for iter_idx in range(max_iter):
        f = _fn(lamb, phi, v, z)
        is_convergence = np.abs(f) <= flooring_fn(0)

        if np.all(is_convergence):
            break

        df = _d_fn(lamb, phi, v, z)
        mu = lamb - f / df
        lamb = np.where(mu > phi_max, mu, (phi_max + lamb) / 2)

    if iter_idx == max_iter - 1:
        f = _fn(lamb, phi, v, z)
        is_convergence = np.abs(f) <= flooring_fn(0)

        if not np.all(is_convergence):
            warnings.warn(
                f"Newton-Raphson method did not converge in {max_iter} iterations.", UserWarning
            )

    if normalization:
        lamb = lamb * phi_max_original

    return lamb


def _find_largest_root(A: np.ndarray, B: np.ndarray, C: np.ndarray) -> np.ndarray:
    r"""Find largest (real) roots of the following cubic equations:

    .. math::

        x^{3} + Ax^{2} + Bx + C = 0.

    Args:
        A (numpy.ndarray): Coefficients of quadratic terms with shape of (\*).
        B (numpy.ndarray): Coefficients of linear terms with shape of (\*).
        C (numpy.ndarray): Coefficients of constant terms with shape of (\*).

    Returns:
        numpy.ndarray of largest (real) roots.

    .. note::

        :math:`x^{3} + Ax^{2} + Bx + C = 0` can be transformed into
        :math:`t^{3} + pt + q = 0` by :math:`t=x+\frac{A}{3}`.
        When :math:`p<0` and :math:`\frac{q^{2}}{4}+\frac{p^{3}}{27}\leq 0`,
        there exists three real solutions: :math:`t=u-\frac{p}{3u}`,
        :math:`x=u\omega-\frac{p\omega^{*}}{3u}`, and
        :math:`x=u\omega^{*}-\frac{p\omega}{3u}`, where

        .. math::

            u
            &=\sqrt[3]{-\frac{q}{2}+\sqrt{\frac{q^{2}}{4} + \frac{p^{3}}{27}}}, \\
            \omega
            &=\frac{-1+j\sqrt{3}}{2}.

        When :math:`p<0` and :math:`\frac{q^{2}}{4}+\frac{p^{3}}{27}>0`,
        :math:`t=u-p/(3u)` is a unique real solution.
        When :math:`p>0`, :math:`t=u-p/(3u)` is a unique real solution.
        Otherwise (when :math:`p=0`), :math:`t=\sqrt[3]{-q}` is a unique real solution.

    """
    P = -(A**2) / 3 + B
    Q = (2 * A**3) / 27 - (A * B) / 3 + C

    omega = (-1 + 1j * np.sqrt(3)) / 2
    omega_conj = (-1 - 1j * np.sqrt(3)) / 2

    discriminant = (Q / 2) ** 2 + (P / 3) ** 3
    discriminant = discriminant.astype(np.complex128)
    U = cbrt(-Q / 2 + np.sqrt(discriminant))
    # When U = 0, P is always 0 in real coefficients cases.
    is_singular = U == 0
    U = np.where(is_singular, 1, U)
    V = -P / (3 * U)

    X1 = U + V
    X1 = np.where(is_singular, cbrt(-Q), X1)
    X2 = np.real(U * omega + V * omega_conj)
    X3 = np.real(U * omega_conj + V * omega)

    roots = np.stack([X1, X2, X3], axis=-1)
    roots = np.real(roots)

    is_monotonic = P >= 0
    is_unique = np.array([True, False, False])

    imaginary_mask = is_monotonic[..., np.newaxis] & ~is_unique
    roots = np.where(imaginary_mask, -float("inf"), roots)
    imaginary_mask = ~is_monotonic[..., np.newaxis] & ~is_unique
    is_positive = discriminant > 0
    roots = np.where(imaginary_mask & is_positive[..., np.newaxis], -float("inf"), roots)
    root = np.max(roots, axis=-1)
    root = root - A / 3

    return root


def _fn(lamb: np.ndarray, phi: np.ndarray, v: np.ndarray, z: np.ndarray) -> np.ndarray:
    r"""Compute values of :math:`f(\lambda_{in})`, where

    .. math::

        f(\lambda_{in})
        = \lambda_{in}^{2}\sum_{n'}
        \frac{\phi_{inn'}|\tilde{v}_{inn'}|^{2}}{(\lambda_{in}-\phi_{inn'})^{2}}
        - \lambda_{in} + z_{in}

    Args:
        lamb (numpy.ndarray): Argument of :math:`f(\lambda_{in})` with shape of (n_bins,).
        phi (numpy.ndarray): Eigen values defined in LQPQM of shape (n_bins, n_sources).
        v (numpy.ndarray): Linear term defined in LQPQM of shape (n_bins, n_sources).
        z (numpy.ndarray): Constant term defined in LQPQM of shape (n_bins,).

    Returns:
        numpy.ndarray of values :math:`f(\lambda_{in})` of shape (n_bins,).

    """
    num = phi * np.abs(v) ** 2
    denom = (lamb[..., np.newaxis] - phi) ** 2
    f = lamb**2 * np.sum(num / denom, axis=-1) - lamb + z

    return f


def _d_fn(
    lamb: np.ndarray,
    phi: np.ndarray,
    v: np.ndarray,
    z: Optional[np.ndarray] = None,
):
    r"""Compute values of :math:`f'(\lambda_{in})`, where

    .. math::

        f'(\lambda_{in})
        = -2\lambda_{in}\sum_{n'}
        \frac{\phi_{inn'}^{2}|\tilde{v}_{inn'}|^{2}}{(\lambda_{in}-\phi_{inn'})^{3}}
        - 1

    Args:
        lamb (numpy.ndarray): Argument of :math:`f'(\lambda_{in})` with shape of (n_bins,).
        phi (numpy.ndarray): Eigen values defined in LQPQM of shape (n_bins, n_sources).
        v (numpy.ndarray): Linear term defined in LQPQM of shape (n_bins, n_sources).
        z (numpy.ndarray, optional): Constant term defined in LQPQM of shape (n_bins,).
            This argument is not used in this funtion.

    Returns:
        numpy.ndarray of values :math:`f'(\lambda_{in})` of shape (n_bins,).

    """
    num = (phi * np.abs(v)) ** 2
    denom = (lamb[..., np.newaxis] - phi) ** 3
    df = -2 * lamb * np.sum(num / denom, axis=-1) - 1

    return df



import numpy as np


def cbrt(x: np.ndarray) -> np.ndarray:
    """Return cube-root of an array.

    Args:
        x (np.ndarray):
            Values to compute cube-root. Complex value is available.

    Returns:
        np.ndarray of cube-root.

    """
    if np.iscomplexobj(x):
        amplitude = np.abs(x)
        phase = np.angle(x)
        x_cbrt = np.cbrt(amplitude) * np.exp(1j * phase / 3)
    else:
        x_cbrt = np.cbrt(x)

    return x_cbrt



import functools
from typing import Callable, Iterable, Optional, Tuple

import numpy as np

from ..linalg import eigh2, inv2
from ..linalg.lqpqm import lqpqm2
from ..special.flooring import identity, max_flooring
from ..special.psd import to_psd
from ..utils.select_pair import sequential_pair_selector

EPS = 1e-10


def update_by_ip1(
    demix_filter: np.ndarray,
    weighted_covariance: np.ndarray,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
    overwrite: bool = True,
) -> np.ndarray:
    r"""Update demixing filters by iterative projection.

    Args:
        demix_filter (numpy.ndarray):
            Demixing filters to be updated.
            The shape is (n_bins, n_sources, n_channels).
        weighted_covariance (numpy.ndarray):
            Weighted covariance matrix.
            The shape is (n_bins, n_sources, n_channels, n_channels).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        overwrite (bool):
            Overwrite ``demix_filter`` if ``overwrite=True``.
            Default: ``True``.

    Returns:
        numpy.ndarray of updated demixing filters.
        The shape is (n_bins, n_sources, n_channels).
    """
    if flooring_fn is None:
        flooring_fn = identity

    if overwrite:
        W = demix_filter
    else:
        W = demix_filter.copy()

    U = weighted_covariance

    n_bins, n_sources, n_channels = W.shape

    E = np.eye(n_sources, n_channels)  # (n_sources, n_channels)
    E = np.tile(E, reps=(n_bins, 1, 1))  # (n_bins, n_sources, n_channels)

    for src_idx in range(n_sources):
        w_n_Hermite = W[:, src_idx, :]  # (n_bins, n_channels)
        U_n = U[:, src_idx, :, :]
        e_n = E[:, src_idx, :]  # (n_bins, n_n_channels)

        WU = W @ U_n
        w_n = np.linalg.solve(WU, e_n)  # (n_bins, n_channels)
        wUw = w_n[:, np.newaxis, :].conj() @ U_n @ w_n[:, :, np.newaxis]
        wUw = np.real(wUw[..., 0])
        wUw = np.maximum(wUw, 0)
        denom = np.sqrt(wUw)
        denom = flooring_fn(denom)
        w_n_Hermite = w_n.conj() / denom
        W[:, src_idx, :] = w_n_Hermite

    return W


def update_by_ip2(
    demix_filter: np.ndarray,
    weighted_covariance: np.ndarray,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
    pair_selector: Optional[Callable[[int], Iterable[Tuple[int, int]]]] = None,
    overwrite: bool = True,
) -> np.ndarray:
    r"""Update demixing filters by pairwise iterative projection [#ono2018fast]_.

    Args:
        demix_filter (numpy.ndarray):
            Demixing filters to be updated.
            The shape is (n_bins, n_sources, n_channels).
        weighted_covariance (numpy.ndarray):
            Weighted covariance matrix.
            The shape is (n_bins, n_sources, n_channels, n_channels).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        pair_selector (callable, optional):
            Selector to choose updaing pair.
            If ``None`` is given, ``sequential_pair_selector`` is used.
            Default: ``None``.
        overwrite (bool):
            Overwrite ``demix_filter`` if ``overwrite=True``.
            Default: ``True``.

    Returns:
        numpy.ndarray of updated demixing filters.
        The shape is (n_bins, n_sources, n_channels).

    .. [#ono2018fast] N. Ono, \
        "Fast algorithm for independent component/vector/low-rank matrix analysis \
        with three or more sources," \
        in *Proc. ASJ Spring meeting*, 2018 (in Japanese).
    """
    if flooring_fn is None:
        flooring_fn = identity

    if pair_selector is None:
        pair_selector = sequential_pair_selector

    if overwrite:
        W = demix_filter
    else:
        W = demix_filter.copy()

    U = weighted_covariance

    _, n_sources, _ = W.shape

    for m, n in pair_selector(n_sources):
        pair = (m, n)
        W[:, pair, :] = update_by_ip2_one_pair(
            W, U[:, pair, :, :], pair=pair, flooring_fn=flooring_fn
        )

    return W


def update_by_iss1(
    separated: np.ndarray,
    weight: np.ndarray,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
) -> np.ndarray:
    r"""Update estimated spectrogram by iterative source steering.

    Args:
        separated (numpy.ndarray):
            Estimated spectrograms to be updated.
            The shape is (n_sources, n_bins, n_frames).
        weight (numpy.ndarray):
            Weights for estimated spectrogram.
            The shape is (n_sources, n_bins, n_frames).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.

    Returns:
        numpy.ndarray of updated spectrograms.
        The shape is (n_sources, n_bins, n_frames).
    """
    if flooring_fn is None:
        flooring_fn = identity

    Y = separated
    varphi = weight

    n_sources = Y.shape[0]

    for src_idx in range(n_sources):
        Y_n = Y[src_idx]  # (n_bins, n_frames)

        YY_n_conj = Y * Y_n.conj()
        YY_n = np.abs(Y_n) ** 2
        num = np.mean(varphi * YY_n_conj, axis=-1)
        denom = np.mean(varphi * YY_n, axis=-1)
        denom = flooring_fn(denom)
        v_n = num / denom
        v_n[src_idx] = 1 - 1 / np.sqrt(denom[src_idx])

        Y = Y - v_n[:, :, np.newaxis] * Y_n

    return Y


def update_by_iss2(
    separated: np.ndarray,
    weight: np.ndarray,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
    pair_selector: Optional[Callable[[int], Iterable[Tuple[int, int]]]] = None,
) -> np.ndarray:
    r"""Update estimated spectrogram by pairwise iterative source steering.

    Args:
        separated (numpy.ndarray):
            Estimated spectrograms to be updated.
            The shape is (n_sources, n_bins, n_frames).
        weight (numpy.ndarray):
            Weights for estimated spectrogram.
            The shape is (n_sources, n_bins, n_frames).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        pair_selector (callable, optional):
            Selector to choose updaing pair.
            If ``None`` is given, ``sequential_pair_selector`` is used.
            Default: ``None``.

    Returns:
        numpy.ndarray of updated spectrograms.
        The shape is (n_sources, n_bins, n_frames).
    """
    Y = separated
    varphi = weight

    n_sources = Y.shape[0]

    if flooring_fn is None:
        flooring_fn = identity

    if pair_selector is None:
        pair_selector = functools.partial(sequential_pair_selector, stop=n_sources, step=2)

    for m, n in pair_selector(n_sources):
        if m < 0:
            m = n_sources + m
        if n < 0:
            n = n_sources + n

        if m > n:
            ascend = False
            m, n = n, m
        else:
            ascend = True

        # Split into main and sub
        Y_1, Y_m, Y_2, Y_n, Y_3 = np.split(Y, [m, m + 1, n, n + 1], axis=0)
        Y_sub = np.concatenate([Y_1, Y_2, Y_3], axis=0)  # (n_sources - 2, n_bins, n_frames)
        varphi_1, varphi_m, varphi_2, varphi_n, varphi_3 = np.split(
            varphi, [m, m + 1, n, n + 1], axis=0
        )
        varphi_sub = np.concatenate([varphi_1, varphi_2, varphi_3], axis=0)

        if ascend:
            Y_main = np.concatenate([Y_m, Y_n], axis=0)  # (2, n_bins, n_frames)
            varphi_main = np.concatenate([varphi_m, varphi_n], axis=0)
        else:
            Y_main = np.concatenate([Y_n, Y_m], axis=0)  # (2, n_bins, n_frames)
            varphi_main = np.concatenate([varphi_n, varphi_m], axis=0)

        YY_main = Y_main[:, np.newaxis, :, :] * Y_main[np.newaxis, :, :, :].conj()
        YY_sub = Y_main[:, np.newaxis, :, :] * Y_sub[np.newaxis, :, :, :].conj()
        YY_main = YY_main.transpose(2, 0, 1, 3)
        YY_sub = YY_sub.transpose(1, 2, 0, 3)

        Y_main = Y_main.transpose(1, 0, 2)

        # Sub
        G_sub = np.mean(
            varphi_sub[:, :, np.newaxis, np.newaxis, :] * YY_main[np.newaxis, :, :, :, :],
            axis=-1,
        )
        F = np.mean(varphi_sub[:, :, np.newaxis, :] * YY_sub, axis=-1)
        Q = -inv2(G_sub) @ F[:, :, :, np.newaxis]
        Q = Q.squeeze(axis=-1)
        Q = Q.transpose(1, 0, 2)
        QY = Q.conj() @ Y_main
        Y_sub = Y_sub + QY.transpose(1, 0, 2)

        # Main
        G_main = np.mean(
            varphi_main[:, :, np.newaxis, np.newaxis, :] * YY_main[np.newaxis, :, :, :, :],
            axis=-1,
        )
        G_m, G_n = G_main
        _, H_mn = eigh2(G_m, G_n)
        h_mn = H_mn.transpose(2, 0, 1)
        hGh_mn = h_mn[:, :, np.newaxis, :].conj() @ G_main @ h_mn[:, :, :, np.newaxis]
        hGh_mn = np.squeeze(hGh_mn, axis=-1)
        hGh_mn = np.real(hGh_mn)
        hGh_mn = np.maximum(hGh_mn, 0)
        denom_mn = np.sqrt(hGh_mn)
        denom_mn = flooring_fn(denom_mn)
        P = h_mn / denom_mn
        P = P.transpose(1, 0, 2)
        Y_main = P.conj() @ Y_main
        Y_main = Y_main.transpose(1, 0, 2)

        # Concat
        Y_m, Y_n = np.split(Y_main, [1], axis=0)
        Y1, Y2, Y3 = np.split(Y_sub, [m, n - 1], axis=0)

        if ascend:
            Y = np.concatenate([Y1, Y_m, Y2, Y_n, Y3], axis=0)
        else:
            Y = np.concatenate([Y1, Y_n, Y2, Y_m, Y3], axis=0)

    return Y


def update_by_ip2_one_pair(
    demix_filter: np.ndarray,
    weighted_covariance_pair: np.ndarray,
    pair: Tuple[int],
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
) -> np.ndarray:
    r"""Update demixing filters by pairwise iterative projection.

    Args:
        demix_filter (numpy.ndarray):
            Demixing filters.
            The shape is (n_bins, n_sources, n_channels).
        weighted_covariance_pair (numpy.ndarray):
            Weighted covariance matrix.
            The shape is (n_bins, 2, n_channels, n_channels).
        pair (tuple):
            Pair of source index to be updated.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.

    Returns:
        numpy.ndarray of updated demixing filter pair.
        The shape is (n_bins, 2, n_channels).
    """
    if flooring_fn is None:
        flooring_fn = identity

    m, n = pair
    W = demix_filter
    U_m, U_n = weighted_covariance_pair.transpose(1, 0, 2, 3)

    n_bins, n_sources, n_channels = W.shape

    E = np.eye(n_channels, n_sources)
    E_mn = E[:, (m, n)]
    E_mn = np.tile(E_mn, reps=(n_bins, 1, 1))

    WU_m = W @ U_m
    WU_n = W @ U_n

    P_m = np.linalg.solve(WU_m, E_mn)
    P_n = np.linalg.solve(WU_n, E_mn)

    PUP_m = P_m.transpose(0, 2, 1).conj() @ U_m @ P_m
    PUP_n = P_n.transpose(0, 2, 1).conj() @ U_n @ P_n

    _, H_mn = eigh2(PUP_m, PUP_n)
    H_mn = H_mn[..., ::-1]

    H_mn = H_mn.transpose(2, 0, 1)
    h_m, h_n = H_mn

    hUh_m = h_m[:, np.newaxis, :].conj() @ PUP_m @ h_m[:, :, np.newaxis]
    hUh_m = np.real(hUh_m[..., 0])
    hUh_m = np.maximum(hUh_m, 0)
    denom = np.sqrt(hUh_m)
    denom = flooring_fn(denom)
    h_m = h_m / denom

    hUh_n = h_n[:, np.newaxis, :].conj() @ PUP_n @ h_n[:, :, np.newaxis]
    hUh_n = np.real(hUh_n[..., 0])
    hUh_n = np.maximum(hUh_n, 0)
    denom = np.sqrt(hUh_n)
    denom = flooring_fn(denom)
    h_n = h_n / denom

    w_m = P_m @ h_m[..., np.newaxis]
    w_n = P_n @ h_n[..., np.newaxis]

    W_mn_conj = np.concatenate([w_m, w_n], axis=-1)
    W_mn = W_mn_conj.transpose(0, 2, 1).conj()

    return W_mn


def update_by_ipa(
    separated: np.ndarray,
    weight: np.ndarray,
    normalization: bool = True,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
    max_iter: int = 1,
) -> np.ndarray:
    r"""Update estimated spectrogram by iterative projection with adjustment (IPA).

    Args:
        separated (numpy.ndarray):
            Estimated spectrograms to be updated.
            The shape is (n_sources, n_bins, n_frames).
        weight (numpy.ndarray):
            Weights for estimated spectrogram.
            The shape is (n_sources, n_bins, n_frames).
        normalization (bool):
            If ``normalization=True``, normalization is applied to LQPQM.
            Default: ``True``.
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.
        max_iter (int):
            Maximum number of Newton-Raphson method. Default: ``1``.

    Returns:
        numpy.ndarray of estimated spectrograms of shape (n_sources, n_bins, n_frames).

    """
    if flooring_fn is None:
        flooring_fn = identity

    Y = separated
    varphi = weight

    n_sources = Y.shape[0]

    E = np.eye(n_sources)

    for source_idx in range(n_sources):
        YY_conj = Y[:, np.newaxis] * Y[np.newaxis, :].conj()
        U_tilde = np.mean(varphi[:, np.newaxis, np.newaxis] * YY_conj, axis=-1)
        U_tilde = U_tilde.transpose(3, 0, 1, 2)
        U_tilde = to_psd(U_tilde, axis1=-2, axis2=-1, flooring_fn=flooring_fn)

        E_n_left, e_n, E_n_right = np.split(E, [source_idx, source_idx + 1], axis=-1)
        E_n = np.concatenate([E_n_left, E_n_right], axis=-1)

        U_tilde_n = U_tilde[:, source_idx, :, :]
        U_tilde_n_inverse = _psd_inv(U_tilde_n, flooring_fn=flooring_fn)
        a_n = U_tilde[:, :, source_idx, source_idx]
        a_n = np.real(a_n)
        a_n = a_n @ E_n
        b_n = np.diagonal(U_tilde[:, :, source_idx, :], axis1=-2, axis2=-1)
        b_n = b_n @ E_n
        d_n = E_n.transpose(1, 0) @ U_tilde_n_inverse.conj()
        C_n = d_n @ E_n
        d_n = d_n[:, :, source_idx]

        Cd_n = np.linalg.solve(C_n, d_n)
        dCd_n = np.sum(d_n.conj() * Cd_n, axis=-1)
        dCd_n = np.real(dCd_n)
        eUe_n = U_tilde_n_inverse[:, source_idx, source_idx]
        eUe_n = np.real(eUe_n)
        z_n = eUe_n - dCd_n

        a_sqrt_n = np.sqrt(a_n)
        aa_n = a_sqrt_n[:, :, np.newaxis] * a_sqrt_n[:, np.newaxis, :]
        H_n = C_n / aa_n
        v_n = -b_n / a_sqrt_n - a_sqrt_n * Cd_n

        if normalization:
            trace = np.trace(H_n, axis1=-2, axis2=-1)
            trace = np.real(trace)

            H_n = H_n / trace[..., np.newaxis, np.newaxis]
            z_n = z_n / trace

        q_check_n = lqpqm2(
            H_n,
            v_n,
            z_n,
            flooring_fn=flooring_fn,
            singular_fn=lambda x: x < flooring_fn(0),
            max_iter=max_iter,
        )

        q_n = q_check_n / a_sqrt_n - b_n / a_n

        Eq_n = q_n.conj() @ E_n.transpose(1, 0)
        q_tilde_n = e_n.transpose(1, 0) - Eq_n

        Uq_n = np.linalg.solve(U_tilde_n, q_tilde_n)
        qUq_n = np.sum(q_tilde_n.conj() * Uq_n, axis=-1, keepdims=True)

        qUq_n = np.real(qUq_n)
        qUq_n = np.maximum(qUq_n, 0)
        denom = np.sqrt(qUq_n)
        denom = flooring_fn(denom)
        p_n = Uq_n / denom

        Y_n = Y[source_idx]
        p_n_conj = p_n.transpose(1, 0).conj()
        PY_n = np.sum(p_n_conj[..., np.newaxis] * Y, axis=0)
        PY_n = e_n[:, np.newaxis] * (PY_n - Y_n)
        Eq_n = Eq_n.transpose(1, 0)
        QY_n = Eq_n[:, :, np.newaxis] * Y_n

        Y = Y + PY_n + QY_n

    return Y


def update_by_block_decomposition_vcd(
    demix_filter: np.ndarray,
    weighted_covariance: np.ndarray,
    singular_fn: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    overwrite: bool = True,
) -> np.ndarray:
    r"""
    Args:
        demix_filter (numpy.ndarray):
            Demixing filters to be updated.
            The shape is (n_blocks, n_neighbors, n_sources, n_channels).
        weighted_covariance (numpy.ndarray):
            Weighted covariance matrix.
            The shape is (n_blocks, n_neighbors, n_neighbors, n_sources, n_channels, n_channels).
        singular_fn (callable, optional):
            A flooring function to return singular condition.
            This function is expected to return the same shape bool tensor as the input.
            If ``singular_fn=None``,``lambda x: x == 0`` is used.
        overwrite (bool):
            Overwrite ``demix_filter`` if ``overwrite=True``.
            Default: ``True``.

    Returns:
        numpy.ndarray of updated demixing filters.
        The shape is (n_blocks, n_neighbors, n_sources, n_channels).
    """
    na = np.newaxis

    if singular_fn is None:

        def _is_zero(x: np.ndarray) -> np.ndarray:
            return x == 0

        singular_fn = _is_zero

    if overwrite:
        W = demix_filter
    else:
        W = demix_filter.copy()

    RXX = weighted_covariance
    U = np.diagonal(RXX, axis1=1, axis2=2)

    n_blocks, n_neighbors, n_sources, n_channels = W.shape

    E_i = np.eye(n_neighbors)
    E_n = np.eye(n_sources)
    E_n = np.tile(E_n, reps=(n_blocks, 1, 1))

    for neighbor_idx in range(n_neighbors):
        pad_mask_i = 1 - E_i[neighbor_idx]

        U_i = U[:, :, :, :, neighbor_idx]
        RXX_i = RXX[:, neighbor_idx]

        for source_idx in range(n_sources):
            e_n = E_n[:, source_idx, :]
            U_in = U_i[:, source_idx, :, :]
            RXX_in = RXX_i[:, :, source_idx]
            w_n_conj = W[:, :, source_idx, :].conj()

            RXY_in = RXX_in @ w_n_conj[:, :, :, na]

            gamma_in = np.sum(pad_mask_i[:, na] * RXY_in[..., 0], axis=1)

            WU_in = W[:, neighbor_idx, :, :] @ U_in
            eta_in = np.linalg.solve(WU_in, e_n)
            eta_hat_in = np.linalg.solve(U_in, gamma_in)
            eta_U_in = eta_in[:, na, :].conj() @ U_in

            xi_in = eta_U_in @ eta_in[:, :, na]
            xi_hat_in = eta_U_in @ eta_hat_in[:, :, na]

            xi_in = np.real(xi_in[..., 0])
            xi_in = np.maximum(xi_in, 0)
            xi_hat_in = xi_hat_in[..., 0]

            singular_condition = singular_fn(xi_hat_in)

            # to avoid zero division, but these will be ignored.
            xi_hat_in[singular_condition] = 1

            coeff = (xi_hat_in / (2 * xi_in)) * (
                1 - np.sqrt(1 + 4 * xi_in / (np.abs(xi_hat_in) ** 2))
            )
            coeff_singular = 1 / np.sqrt(xi_in)
            coeff = np.where(singular_condition, coeff_singular, coeff)

            w_in = coeff * eta_in - eta_hat_in

            W[:, neighbor_idx, source_idx, :] = w_in.conj()

    return W


def _psd_inv(
    X: np.ndarray,
    flooring_fn: Optional[Callable[[np.ndarray], np.ndarray]] = functools.partial(
        max_flooring, eps=EPS
    ),
) -> np.ndarray:
    """Compute inversion of positive semidefinite matrix.

    Args:
        X (np.ndarray): Positive semidefinite matrix of shape (*, N, N).
        flooring_fn (callable, optional):
            A flooring function for numerical stability.
            This function is expected to return the same shape tensor as the input.
            If you explicitly set ``flooring_fn=None``,
            the identity function (``lambda x: x``) is used.
            Default: ``functools.partial(max_flooring, eps=1e-10)``.

    Returns:
        np.ndarray: Inversion of input matrix.

    """
    if flooring_fn is None:
        flooring_fn = identity

    Lamb, P = np.linalg.eigh(X)

    P_Hermite = P.swapaxes(-2, -1)

    if np.iscomplexobj(X):
        P_Hermite = P_Hermite.conj()

    Lamb_inv = 1 / flooring_fn(Lamb)
    Lamb_inv = Lamb_inv[..., np.newaxis] * np.eye(Lamb.shape[-1])

    return P @ Lamb_inv @ P_Hermite



import warnings
from typing import Callable, List, Optional, Union

import numpy as np

from ..linalg import prox
from .proxbss import ProxBSSBase

EPS = 1e-10

__all__ = ["PDSBSS", "MaskingPDSBSS"]


class PDSBSSBase(ProxBSSBase):
    r"""Base class of blind source separation \
    via proximal splitting algorithm [#yatabe2018determined]_.

    Args:
        penalty_fn (callable):
            Penalty function that determines source model.
        prox_penalty (callable):
            Proximal operator of penalty function.
            Default: ``None``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.

    .. [#yatabe2018determined] K. Yatabe and D. Kitamura,
        "Determined blind source separation via proximal splitting algorithm,"
        in *Proc. ICASSP*, 2018, pp. 776-780.
    """

    def __repr__(self) -> str:
        s = "PDSBSS("
        s += "n_penalties={n_penalties}".format(n_penalties=self.n_penalties)
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)


class PDSBSS(PDSBSSBase):
    r"""Blind source separation via proximal splitting algorithm [#yatabe2018determined]_.

    Args:
        mu1 (float):
            Step size. Default: ``1``.
        mu2 (float):
            Step size. Default: ``1``.
        alpha (float):
            Relaxation parameter (deprecated). Set ``relaxation`` instead.
        relaxation (float):
            Relaxation parameter. Default: ``1``.
        penalty_fn (callable, optional):
            Penalty function that determines source model.
        prox_penalty (callable):
            Proximal operator of penalty function.
            Default: ``None``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool, optional):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``None``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.
    """

    def __init__(
        self,
        mu1: float = 1,
        mu2: float = 1,
        alpha: float = None,
        relaxation: float = 1,
        penalty_fn: Optional[Callable[[np.ndarray, np.ndarray], float]] = None,
        prox_penalty: Callable[[np.ndarray, float], np.ndarray] = None,
        callbacks: Optional[
            Union[Callable[["PDSBSS"], None], List[Callable[["PDSBSS"], None]]]
        ] = None,
        scale_restoration: bool = True,
        record_loss: Optional[bool] = None,
        reference_id: int = 0,
    ) -> None:
        super().__init__(
            penalty_fn=penalty_fn,
            prox_penalty=prox_penalty,
            callbacks=callbacks,
            scale_restoration=scale_restoration,
            record_loss=record_loss,
            reference_id=reference_id,
        )

        self.mu1, self.mu2 = mu1, mu2

        if alpha is None:
            self.relaxation = relaxation
        else:
            assert relaxation == 1, "You cannot specify relaxation and alpha simultaneously."

            warnings.warn("alpha is deprecated. Set relaxation instead.", DeprecationWarning)

            self.relaxation = alpha

    def __call__(self, input, n_iter=100, initial_call: bool = True, **kwargs) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                Mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                Number of iterations of demixing filter updates.
                Default: ``100``.
            initial_call (bool):
                If ``True``, perform callbacks (and computation of loss if necessary)
                before iterations.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        # Call __call__ of PDSBSSBase's parent, i.e. __call__ of IterativeMethodBase
        super(PDSBSSBase, self).__call__(n_iter=n_iter, initial_call=initial_call)

        if self.scale_restoration:
            self.restore_scale()

        self.output = self.separate(self.input, demix_filter=self.demix_filter)

        return self.output

    def __repr__(self) -> str:
        s = "PDSBSS("
        s += "mu1={mu1}, mu2={mu2}"
        s += ", relaxation={relaxation}"
        s += ", n_penalties={n_penalties}".format(n_penalties=self.n_penalties)
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(self, **kwargs) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            kwargs:
                Keyword arguments to set as attributes of PDSBSS.
        """
        super()._reset(**kwargs)

        n_penalties = self.n_penalties
        n_sources = self.n_sources
        n_bins, n_frames = self.n_bins, self.n_frames

        if not hasattr(self, "dual"):
            dual = np.zeros((n_penalties, n_sources, n_bins, n_frames), dtype=np.complex128)
        else:
            if self.dual is None:
                dual = None
            else:
                # To avoid overwriting ``dual`` given by keyword arguments.
                dual = self.dual.copy()

        self.dual = dual

    def update_once(self) -> None:
        r"""Update demixing filters and dual parameters once."""
        mu1, mu2 = self.mu1, self.mu2
        alpha = self.relaxation

        Y = self.dual
        X, W = self.input, self.demix_filter

        Y_sum = Y.sum(axis=0)
        XY = Y_sum.transpose(1, 0, 2) @ X.transpose(1, 2, 0).conj()
        W_tilde = prox.neg_logdet(W - mu1 * mu2 * XY, step_size=mu1)
        XW = self.separate(X, demix_filter=2 * W_tilde - W)
        Y_tilde = []

        for Y_q, prox_penalty in zip(Y, self.prox_penalty):
            Z_q = Y_q + XW
            Y_tilde_q = Z_q - prox_penalty(Z_q, step_size=1 / mu2)
            Y_tilde.append(Y_tilde_q)

        Y_tilde = np.stack(Y_tilde, axis=0)

        self.demix_filter = alpha * W_tilde + (1 - alpha) * W
        self.dual = alpha * Y_tilde + (1 - alpha) * Y


class MaskingPDSBSS(PDSBSSBase):
    r"""Blind source separation via proximal splitting algorithm with masking [#yatabe2019time]_.

    Args:
        mu1 (float):
            Step size. Default: ``1``.
        mu2 (float):
            Step size. Default: ``1``.
        alpha (float):
            Relaxation parameter (deprecated). Set ``relaxation`` instead.
        relaxation (float):
            Relaxation parameter. Default: ``1``.
        penalty_fn (callable, optional):
            Penalty function that determines source model.
        mask_fn (callable):
            Proximal operator of penalty function.
            Default: ``None``.
        callbacks (callable or list[callable], optional):
            Callback functions. Each function is called before separation and at each iteration.
            Default: ``None``.
        scale_restoration (bool or str):
            Technique to restore scale ambiguity.
            If ``scale_restoration=True``, the projection back technique is applied to
            estimated spectrograms. You can also specify ``projection_back`` explicitly.
            Default: ``True``.
        record_loss (bool):
            Record the loss at each iteration of the update algorithm if ``record_loss=True``.
            Default: ``True``.
        reference_id (int):
            Reference channel for projection back.
            Default: ``0``.

    .. [#yatabe2019time] K. Yatabe and D. Kitamura,
        "Time-frequency-masking-based determined BSS with application to sparse IVA,"
        in *Proc. ICASSP*, pp. 715-719, 2019.

    """

    def __init__(
        self,
        mu1: float = 1,
        mu2: float = 1,
        alpha: float = None,
        relaxation: float = 1,
        penalty_fn: Optional[Callable[[np.ndarray, np.ndarray], float]] = None,
        mask_fn: Callable[[np.ndarray], float] = None,
        callbacks: Optional[
            Union[Callable[["MaskingPDSBSS"], None], List[Callable[["MaskingPDSBSS"], None]]]
        ] = None,
        scale_restoration: bool = True,
        record_loss: Optional[bool] = None,
        reference_id: int = 0,
    ) -> None:
        super(ProxBSSBase, self).__init__(
            callbacks=callbacks,
            record_loss=record_loss,
        )

        if penalty_fn is None:
            # Since penalty_fn is not necessarily written in closed form,
            # None is acceptable.
            if record_loss is None:
                record_loss = False

            assert not record_loss, "To record loss, set penalty_fn."
        else:
            assert callable(penalty_fn), "penalty_fn should be callable."

            if record_loss is None:
                record_loss = True

        if mask_fn is None:
            raise ValueError("Specify masking function.")
        else:
            assert callable(mask_fn), "mask_fn should be callable."

        self.penalty_fn = penalty_fn
        self.mask_fn = mask_fn

        self.input = None
        self.scale_restoration = scale_restoration

        if reference_id is None and scale_restoration:
            raise ValueError("Specify 'reference_id' if scale_restoration=True.")
        else:
            self.reference_id = reference_id

        self.mu1, self.mu2 = mu1, mu2

        if alpha is None:
            self.relaxation = relaxation
        else:
            assert relaxation == 1, "You cannot specify relaxation and alpha simultaneously."

            warnings.warn("alpha is deprecated. Set relaxation instead.", DeprecationWarning)

            self.relaxation = alpha

    def __call__(self, input, n_iter=100, initial_call: bool = True, **kwargs) -> np.ndarray:
        r"""Separate a frequency-domain multichannel signal.

        Args:
            input (numpy.ndarray):
                Mixture signal in frequency-domain.
                The shape is (n_channels, n_bins, n_frames).
            n_iter (int):
                Number of iterations of demixing filter updates.
                Default: ``100``.
            initial_call (bool):
                If ``True``, perform callbacks (and computation of loss if necessary)
                before iterations.

        Returns:
            numpy.ndarray of the separated signal in frequency-domain.
            The shape is (n_channels, n_bins, n_frames).
        """
        self.input = input.copy()

        self._reset(**kwargs)

        # Call __call__ of PDSBSSBase's parent, i.e. __call__ of IterativeMethodBase
        super(PDSBSSBase, self).__call__(n_iter=n_iter, initial_call=initial_call)

        if self.scale_restoration:
            self.restore_scale()

        self.output = self.separate(self.input, demix_filter=self.demix_filter)

        return self.output

    def __repr__(self) -> str:
        s = "MaskingPDSBSS("
        s += "mu1={mu1}, mu2={mu2}"
        s += ", relaxation={relaxation}"
        s += ", scale_restoration={scale_restoration}"
        s += ", record_loss={record_loss}"

        if self.scale_restoration:
            s += ", reference_id={reference_id}"

        s += ")"

        return s.format(**self.__dict__)

    def _reset(self, **kwargs) -> None:
        r"""Reset attributes by given keyword arguments.

        Args:
            kwargs:
                Keyword arguments to set as attributes of MaskingPDSBSS.
        """
        super()._reset(**kwargs)

        assert self.n_penalties == 1, "Number of penalty function should be one."

        n_sources = self.n_sources
        n_bins, n_frames = self.n_bins, self.n_frames

        if not hasattr(self, "dual"):
            dual = np.zeros((n_sources, n_bins, n_frames), dtype=np.complex128)
        else:
            if self.dual is None:
                dual = None
            else:
                # To avoid overwriting ``dual`` given by keyword arguments.
                dual = self.dual.copy()

        self.dual = dual

    @property
    def n_penalties(self):
        r"""Return number of penalty terms."""
        return 1

    def update_once(self) -> None:
        r"""Update demixing filters and dual parameters once."""
        mu1, mu2 = self.mu1, self.mu2
        alpha = self.relaxation

        Y = self.dual
        X, W = self.input, self.demix_filter

        XY = Y.transpose(1, 0, 2) @ X.transpose(1, 2, 0).conj()
        W_tilde = prox.neg_logdet(W - mu1 * mu2 * XY, step_size=mu1)
        XW = self.separate(X, demix_filter=2 * W_tilde - W)

        Z = Y + XW
        Y_tilde = Z - self.mask_fn(Z) * Z

        self.demix_filter = alpha * W_tilde + (1 - alpha) * W
        self.dual = alpha * Y_tilde + (1 - alpha) * Y
